"""UI tests for the appalerts app"""

import time

from django.urls import reverse

from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as ec
from selenium.webdriver.support.ui import WebDriverWait

# Projectroles dependency
from projectroles.app_settings import AppSettingAPI
from projectroles.tests.test_ui import UITestBase

from appalerts.models import AppAlert
from appalerts.tests.test_models import AppAlertMixin


app_settings = AppSettingAPI()


# Local constants
APP_NAME_PR = 'projectroles'


class AlertUITestBase(AppAlertMixin, UITestBase):
    def setUp(self):
        super().setUp()
        # Create users
        self.superuser = self.make_user('superuser', superuser=True)
        self.superuser.is_superuser = True
        self.superuser.is_staff = True
        self.superuser.save()
        self.regular_user = self.make_user('regular_user')
        self.no_alert_user = self.make_user('no_alert_user')
        # No user
        self.anonymous = None
        # Create alerts
        self.alert = self.make_app_alert(
            user=self.regular_user, url=reverse('home')
        )
        self.alert2 = self.make_app_alert(
            user=self.regular_user, url=reverse('home')
        )


class TestAppAlertListView(AlertUITestBase):
    """Tests for AppAlertListView"""

    def _find_alert_element(self, alert: AppAlert) -> WebElement:
        """Return element for AppAlert object"""
        return self.selenium.find_element(
            By.XPATH, f'//div[@data-alert-uuid="{alert.sodar_uuid}"]'
        )

    def setUp(self):
        super().setUp()
        self.url = reverse('appalerts:list')
        self.url_dismiss = reverse('appalerts:list_dismissed')

    def test_render(self):
        """Test existence of alert items in list"""
        expected = [
            (self.superuser, 0),
            (self.regular_user, 2),
            (self.no_alert_user, 0),
        ]
        self.assert_element_count(
            expected, self.url, 'sodar-app-alert-item', 'class'
        )

    def test_render_links(self):
        """Test rendering alert dismiss links"""
        self.login_and_redirect(self.regular_user, self.url)
        elem = self.selenium.find_element(By.ID, 'sodar-ap-ops-dismiss-all')
        self.assertIsNone(elem.get_attribute('disabled'))
        elem = self.selenium.find_element(By.CLASS_NAME, 'sodar-ap-btn-link')
        self.assertIsNone(elem.get_attribute('disabled'))
        elem = self.selenium.find_element(
            By.CLASS_NAME, 'sodar-ap-btn-dismiss-single'
        )
        self.assertIsNone(elem.get_attribute('disabled'))

    def test_render_links_read_only(self):
        """Test rendering alert dismiss links with site read-only mode"""
        app_settings.set(APP_NAME_PR, 'site_read_only', True)
        self.login_and_redirect(self.regular_user, self.url)
        elem = self.selenium.find_element(By.ID, 'sodar-ap-ops-dismiss-all')
        self.assertEqual(elem.get_attribute('disabled'), 'true')
        with self.assertRaises(NoSuchElementException):
            self.selenium.find_element(By.CLASS_NAME, 'sodar-ap-btn-link')
        with self.assertRaises(NoSuchElementException):
            self.selenium.find_element(
                By.CLASS_NAME, 'sodar-ap-btn-dismiss-single'
            )

    def test_render_links_read_only_superuser(self):
        """Test rendering links with site read-only mode as superuser"""
        self.alert.user = self.superuser
        self.alert.save()
        self.alert2.user = self.superuser
        self.alert2.save()
        app_settings.set(APP_NAME_PR, 'site_read_only', True)
        self.login_and_redirect(self.superuser, self.url)
        elem = self.selenium.find_element(By.ID, 'sodar-ap-ops-dismiss-all')
        self.assertIsNone(elem.get_attribute('disabled'))
        elem = self.selenium.find_element(By.CLASS_NAME, 'sodar-ap-btn-link')
        self.assertIsNone(elem.get_attribute('disabled'))
        elem = self.selenium.find_element(
            By.CLASS_NAME, 'sodar-ap-btn-dismiss-single'
        )
        self.assertIsNone(elem.get_attribute('disabled'))

    def test_render_dismissed(self):
        """Test displaying dismissed alerts"""
        self.alert.active = False
        self.alert.save()
        self.alert2.active = False
        self.alert2.save()
        self.login_and_redirect(self.regular_user, self.url_dismiss)
        self.assertIsNotNone(
            self.selenium.find_element(By.ID, 'sodar-ap-btn-active')
        )
        self.assertIsNotNone(self._find_alert_element(self.alert))
        self.assertIsNotNone(self._find_alert_element(self.alert2))

    def test_render_dismissed_project(self):
        """Test displaying dismissed alerts with project"""
        self.make_assignment(
            self.project, self.regular_user, self.role_contributor
        )
        self.alert.project = self.project
        self.alert.active = False
        self.alert.save()
        self.login_and_redirect(self.regular_user, self.url_dismiss)
        alert_cat = self._find_alert_element(self.alert)
        self.assertIn('alert-info', alert_cat.get_attribute('class'))

    def test_render_dismissed_project_no_role(self):
        """Test displaying dismissed alerts with project and no access"""
        # No role assignment created
        self.alert.project = self.project
        self.alert.active = False
        self.alert.save()
        self.login_and_redirect(self.regular_user, self.url_dismiss)
        alert_cat = self._find_alert_element(self.alert)
        self.assertIn('alert-secondary', alert_cat.get_attribute('class'))

    def test_alert_dismiss(self):
        """Test dismissing alert"""
        self.assertEqual(AppAlert.objects.filter(active=True).count(), 2)
        self.login_and_redirect(self.regular_user, self.url)
        self.assertEqual(
            self.selenium.find_element(By.ID, 'sodar-app-alert-count').text, '2'
        )
        self.assertEqual(
            self.selenium.find_element(By.ID, 'sodar-app-alert-legend').text,
            'alerts',
        )

        button = self.selenium.find_elements(
            By.CLASS_NAME, 'sodar-ap-btn-dismiss-single'
        )[0]
        button.click()
        WebDriverWait(self.selenium, self.wait_time).until(
            ec.invisibility_of_element_located(
                (By.CLASS_NAME, 'sodar-app-alert-item')
            )
        )
        self.assertEqual(
            self.selenium.find_element(By.ID, 'sodar-app-alert-count').text, '1'
        )
        self.assertEqual(
            self.selenium.find_element(By.ID, 'sodar-app-alert-legend').text,
            'alert',
        )
        self.assertEqual(AppAlert.objects.filter(active=True).count(), 1)
        self.assertFalse(
            self.selenium.find_element(
                By.ID, 'sodar-app-alert-empty'
            ).is_displayed()
        )

    def test_alert_dismiss_all(self):
        """Test dismissing all alerts for user"""
        self.assertEqual(AppAlert.objects.filter(active=True).count(), 2)
        self.login_and_redirect(self.regular_user, self.url)
        self.assertEqual(
            self.selenium.find_element(By.ID, 'sodar-app-alert-count').text, '2'
        )
        self.assertEqual(
            self.selenium.find_element(By.ID, 'sodar-app-alert-legend').text,
            'alerts',
        )

        self.selenium.find_element(By.ID, 'sodar-ap-ops-dropdown').click()
        WebDriverWait(self.selenium, self.wait_time).until(
            ec.element_to_be_clickable((By.ID, 'sodar-ap-ops-dismiss-all'))
        )
        self.selenium.find_element(By.ID, 'sodar-ap-ops-dismiss-all').click()
        WebDriverWait(self.selenium, self.wait_time).until(
            ec.invisibility_of_element_located(
                (By.CLASS_NAME, 'sodar-app-alert-item')
            )
        )
        self.assertEqual(
            self.selenium.find_element(By.ID, 'sodar-app-alert-count').text, ''
        )
        self.assertEqual(
            self.selenium.find_element(By.ID, 'sodar-app-alert-legend').text,
            '',
        )
        self.assertEqual(AppAlert.objects.filter(active=True).count(), 0)
        WebDriverWait(self.selenium, self.wait_time).until(
            ec.visibility_of_element_located((By.ID, 'sodar-app-alert-empty'))
        )
        self.assertTrue(
            self.selenium.find_element(
                By.ID, 'sodar-app-alert-empty'
            ).is_displayed()
        )

    def test_alert_reload(self):
        """Test displaying reload link for new alerts"""
        AppAlert.objects.filter(active=True).delete()
        self.assertEqual(AppAlert.objects.filter(active=True).count(), 0)
        self.login_and_redirect(self.regular_user, self.url)
        self.assertTrue(
            self.selenium.find_element(
                By.ID, 'sodar-app-alert-empty'
            ).is_displayed()
        )
        with self.assertRaises(NoSuchElementException):
            self.selenium.find_element(By.ID, 'sodar-app-alert-reload')

        self.make_app_alert(user=self.regular_user, url=reverse('home'))

        WebDriverWait(self.selenium, self.wait_time).until(
            ec.visibility_of_element_located((By.ID, 'sodar-app-alert-reload'))
        )
        self.assertTrue(
            self.selenium.find_element(
                By.ID, 'sodar-app-alert-reload'
            ).is_displayed()
        )


class TestTitlebarBadge(AlertUITestBase):
    """Tests for the site titlebar badge"""

    def setUp(self):
        super().setUp()
        self.url = reverse('home')

    def test_render(self):
        """Test existence of alert badge for user with alerts"""
        self.login_and_redirect(self.regular_user, self.url)
        alert_badge = self.selenium.find_element(By.ID, 'sodar-app-alert-badge')
        self.assertIsNotNone(alert_badge)
        self.assertTrue(alert_badge.is_displayed())
        alert_link = self.selenium.find_element(
            By.ID, 'sodar-app-alert-badge-dismiss'
        )
        self.assertIsNotNone(alert_link)
        self.assertTrue(alert_link.is_displayed())
        alert_count = self.selenium.find_element(By.ID, 'sodar-app-alert-count')
        alert_legend = self.selenium.find_element(
            By.ID, 'sodar-app-alert-legend'
        )
        self.assertEqual(alert_count.text, '2')
        self.assertEqual(alert_legend.text, 'alerts')

    def test_render_no_alerts(self):
        """Test existence of alert badge for user without alerts"""
        self.login_and_redirect(self.no_alert_user, self.url)
        alert_badge = self.selenium.find_element(By.ID, 'sodar-app-alert-badge')
        self.assertIsNotNone(alert_badge)
        self.assertFalse(alert_badge.is_displayed())

    def test_render_add(self):
        """Test adding an alert for user with alerts"""
        self.login_and_redirect(self.regular_user, self.url)
        alert_count = self.selenium.find_element(By.ID, 'sodar-app-alert-count')
        alert_legend = self.selenium.find_element(
            By.ID, 'sodar-app-alert-legend'
        )
        self.assertEqual(alert_count.text, '2')
        self.assertEqual(alert_legend.text, 'alerts')

        self.make_app_alert(user=self.regular_user, url=reverse('home'))
        WebDriverWait(self.selenium, self.wait_time).until(
            ec.text_to_be_present_in_element(
                (By.ID, 'sodar-app-alert-count'), '3'
            )
        )
        self.assertEqual(alert_count.text, '3')
        self.assertEqual(alert_legend.text, 'alerts')

    def test_render_delete(self):
        """Test deleting an alert from user with alerts"""
        self.login_and_redirect(self.regular_user, self.url)
        alert_count = self.selenium.find_element(By.ID, 'sodar-app-alert-count')
        alert_legend = self.selenium.find_element(
            By.ID, 'sodar-app-alert-legend'
        )
        self.assertEqual(alert_count.text, '2')
        self.assertEqual(alert_legend.text, 'alerts')

        self.alert2.delete()
        WebDriverWait(self.selenium, self.wait_time).until(
            ec.text_to_be_present_in_element(
                (By.ID, 'sodar-app-alert-count'), '1'
            )
        )
        self.assertEqual(alert_count.text, '1')
        self.assertEqual(alert_legend.text, 'alert')

    def test_render_delete_all(self):
        """Test deleting all alerts from user with alerts"""
        self.login_and_redirect(self.regular_user, self.url)
        alert_badge = self.selenium.find_element(By.ID, 'sodar-app-alert-badge')
        self.assertTrue(alert_badge.is_displayed())

        self.alert.delete()
        self.alert2.delete()
        WebDriverWait(self.selenium, self.wait_time).until(
            ec.invisibility_of_element_located((By.ID, 'sodar-app-alert-badge'))
        )
        self.assertFalse(alert_badge.is_displayed())

    def test_render_add_no_alerts(self):
        """Test adding an alert for user without prior alerts"""
        self.login_and_redirect(self.no_alert_user, self.url)
        alert_badge = self.selenium.find_element(By.ID, 'sodar-app-alert-badge')
        self.assertFalse(alert_badge.is_displayed())

        self.make_app_alert(user=self.no_alert_user, url=reverse('home'))
        WebDriverWait(self.selenium, self.wait_time).until(
            ec.visibility_of_element_located((By.ID, 'sodar-app-alert-badge'))
        )
        self.assertTrue(alert_badge.is_displayed())

        alert_count = self.selenium.find_element(By.ID, 'sodar-app-alert-count')
        alert_legend = self.selenium.find_element(
            By.ID, 'sodar-app-alert-legend'
        )
        self.assertEqual(alert_count.text, '1')
        self.assertEqual(alert_legend.text, 'alert')

    def test_alert_dismiss_all(self):
        """Test dismissing all alerts for the user"""
        self.assertEqual(AppAlert.objects.filter(active=True).count(), 2)
        self.login_and_redirect(self.regular_user, self.url)
        self.selenium.find_element(
            By.ID, 'sodar-app-alert-badge-dismiss'
        ).click()
        WebDriverWait(self.selenium, self.wait_time).until(
            ec.invisibility_of_element_located(
                (By.CLASS_NAME, 'sodar-app-alert-legend')
            )
        )
        time.sleep(2)  # HACK: Timing issue, must wait just in case
        self.assertEqual(AppAlert.objects.filter(active=True).count(), 0)

    def test_render_read_only(self):
        """Test rendering alert badge with site read-only mode"""
        app_settings.set(APP_NAME_PR, 'site_read_only', True)
        self.login_and_redirect(self.regular_user, self.url)
        alert_badge = self.selenium.find_element(By.ID, 'sodar-app-alert-badge')
        self.assertIsNotNone(alert_badge)
        self.assertTrue(alert_badge.is_displayed())
        # Link should not be displayed
        with self.assertRaises(NoSuchElementException):
            self.selenium.find_element(By.ID, 'sodar-app-alert-badge-dismiss')

    def test_render_read_only_superuser(self):
        """Test rendering alert badge with site read-only mode as superuser"""
        app_settings.set(APP_NAME_PR, 'site_read_only', True)
        self.alert.user = self.superuser
        self.alert.save()
        self.login_and_redirect(self.superuser, self.url)
        alert_badge = self.selenium.find_element(By.ID, 'sodar-app-alert-badge')
        self.assertIsNotNone(alert_badge)
        self.assertTrue(alert_badge.is_displayed())
        alert_link = self.selenium.find_element(
            By.ID, 'sodar-app-alert-badge-dismiss'
        )
        self.assertIsNotNone(alert_link)



"""Tests for plugins in the projectroles Django app"""

from test_plus.test import TestCase

from projectroles.models import SODAR_CONSTANTS
from projectroles.plugins import PluginAppSettingDef

# SODAR constants
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
APP_SETTING_SCOPE_PROJECT = SODAR_CONSTANTS['APP_SETTING_SCOPE_PROJECT']
APP_SETTING_SCOPE_USER = SODAR_CONSTANTS['APP_SETTING_SCOPE_USER']
APP_SETTING_SCOPE_PROJECT_USER = SODAR_CONSTANTS[
    'APP_SETTING_SCOPE_PROJECT_USER'
]
APP_SETTING_SCOPE_SITE = SODAR_CONSTANTS['APP_SETTING_SCOPE_SITE']
APP_SETTING_TYPE_BOOLEAN = SODAR_CONSTANTS['APP_SETTING_TYPE_BOOLEAN']
APP_SETTING_TYPE_INTEGER = SODAR_CONSTANTS['APP_SETTING_TYPE_INTEGER']
APP_SETTING_TYPE_JSON = SODAR_CONSTANTS['APP_SETTING_TYPE_JSON']
APP_SETTING_TYPE_STRING = SODAR_CONSTANTS['APP_SETTING_TYPE_STRING']

# Local constants
DEF_SCOPE_INVALID = 'INVALID_SCOPE'
DEF_TYPE_INVALID = 'INVALID_TYPE'
DEF_NAME = 'test_app_setting'
DEF_LABEL = 'Label'
DEF_PLACEHOLDER = 'placeholder'
DEF_DESC = 'description'
DEF_WIDGET_ATTRS = {'class': 'text-danger'}
DEF_JSON_VALUE = {
    'Example': 'Value',
    'list': [1, 2, 3, 4, 5],
    'level_6': False,
}


class TestPluginAppSettingDef(TestCase):
    """Tests for PluginAppSettingDef"""

    def test_init(self):
        """Test PluginAppSettingDef initialization"""
        s_def = PluginAppSettingDef(
            name=DEF_NAME,
            scope=APP_SETTING_SCOPE_PROJECT,
            type=APP_SETTING_TYPE_STRING,
        )
        expected = {
            'name': DEF_NAME,
            'scope': APP_SETTING_SCOPE_PROJECT,
            'type': APP_SETTING_TYPE_STRING,
            'default': '',
            'label': None,
            'placeholder': None,
            'description': None,
            'options': [],
            'user_modifiable': True,
            'global_edit': False,
            'project_types': [PROJECT_TYPE_PROJECT],
            'widget_attrs': {},
        }
        self.assertEqual(s_def.__dict__, expected)

    def test_init_no_defaults(self):
        """Test init with no default values"""
        s_def = PluginAppSettingDef(
            name=DEF_NAME,
            scope=APP_SETTING_SCOPE_PROJECT,
            type=APP_SETTING_TYPE_STRING,
            default='string',
            label=DEF_LABEL,
            placeholder=DEF_PLACEHOLDER,
            description=DEF_DESC,
            options=['string', 'string2'],
            user_modifiable=False,
            global_edit=True,
            project_types=[PROJECT_TYPE_PROJECT, PROJECT_TYPE_CATEGORY],
            widget_attrs=DEF_WIDGET_ATTRS,
        )
        expected = {
            'name': DEF_NAME,
            'scope': APP_SETTING_SCOPE_PROJECT,
            'type': APP_SETTING_TYPE_STRING,
            'default': 'string',
            'label': DEF_LABEL,
            'placeholder': DEF_PLACEHOLDER,
            'description': DEF_DESC,
            'options': ['string', 'string2'],
            'user_modifiable': False,
            'global_edit': True,
            'project_types': [PROJECT_TYPE_PROJECT, PROJECT_TYPE_CATEGORY],
            'widget_attrs': DEF_WIDGET_ATTRS,
        }
        self.assertEqual(s_def.__dict__, expected)

    def test_init_option_default_tuple(self):
        """Test init with option label tuples and default value"""
        s_def = PluginAppSettingDef(
            name=DEF_NAME,
            scope=APP_SETTING_SCOPE_PROJECT,
            type=APP_SETTING_TYPE_STRING,
            default='string2',
            options=['string', ('string2', 'string2 label')],
        )
        self.assertIsNotNone(s_def)

    def test_init_invalid_scope(self):
        """Test init with invalid scope"""
        with self.assertRaises(ValueError):
            PluginAppSettingDef(
                name=DEF_NAME,
                scope=DEF_SCOPE_INVALID,
                type=APP_SETTING_TYPE_STRING,
            )

    def test_init_invalid_type(self):
        """Test init with invalid type"""
        with self.assertRaises(ValueError):
            PluginAppSettingDef(
                name=DEF_NAME,
                scope=APP_SETTING_SCOPE_PROJECT,
                type=DEF_TYPE_INVALID,
            )

    def test_init_invalid_option_type(self):
        """Test init with invalid option type"""
        with self.assertRaises(ValueError):
            PluginAppSettingDef(
                name=DEF_NAME,
                scope=APP_SETTING_SCOPE_PROJECT,
                type=APP_SETTING_TYPE_BOOLEAN,
                options=['string', 'string2'],
            )

    def test_init_default_boolean(self):
        """Test init with BOOLEAN type and valid default"""
        s_def = PluginAppSettingDef(
            name=DEF_NAME,
            scope=APP_SETTING_SCOPE_PROJECT,
            type=APP_SETTING_TYPE_BOOLEAN,
            default=True,
        )
        self.assertIsNotNone(s_def)

    def test_init_default_boolean_invalid(self):
        """Test init with BOOLEAN type and invalid default"""
        with self.assertRaises(ValueError):
            PluginAppSettingDef(
                name=DEF_NAME,
                scope=APP_SETTING_SCOPE_PROJECT,
                type=APP_SETTING_TYPE_BOOLEAN,
                default='abc',
            )

    def test_init_default_integer(self):
        """Test init with INTEGER type and valid default"""
        s_def = PluginAppSettingDef(
            name=DEF_NAME,
            scope=APP_SETTING_SCOPE_PROJECT,
            type=APP_SETTING_TYPE_INTEGER,
            default=0,
        )
        self.assertIsNotNone(s_def)

    def test_init_default_integer_as_string(self):
        """Test init with INTEGER type and valid default as string"""
        s_def = PluginAppSettingDef(
            name=DEF_NAME,
            scope=APP_SETTING_SCOPE_PROJECT,
            type=APP_SETTING_TYPE_INTEGER,
            default='0',
        )
        self.assertIsNotNone(s_def)

    def test_init_default_integer_invalid(self):
        """Test init with INTEGER type and invalid default"""
        with self.assertRaises(ValueError):
            PluginAppSettingDef(
                name=DEF_NAME,
                scope=APP_SETTING_SCOPE_PROJECT,
                type=APP_SETTING_TYPE_INTEGER,
                default='abc',
            )

    def test_init_default_json(self):
        """Test init with JSON type and valid default"""
        s_def = PluginAppSettingDef(
            name=DEF_NAME,
            scope=APP_SETTING_SCOPE_PROJECT,
            type=APP_SETTING_TYPE_JSON,
            default=DEF_JSON_VALUE,
        )
        self.assertIsNotNone(s_def)

    def test_init_default_json_empty_dict(self):
        """Test init with JSON type and valid empty dict default"""
        s_def = PluginAppSettingDef(
            name=DEF_NAME,
            scope=APP_SETTING_SCOPE_PROJECT,
            type=APP_SETTING_TYPE_JSON,
            default={},
        )
        self.assertIsNotNone(s_def)

    def test_init_default_json_empty_list(self):
        """Test init with JSON type and valid empty list default"""
        s_def = PluginAppSettingDef(
            name=DEF_NAME,
            scope=APP_SETTING_SCOPE_PROJECT,
            type=APP_SETTING_TYPE_JSON,
            default=[],
        )
        self.assertIsNotNone(s_def)

    def test_init_default_json_invalid(self):
        """Test init with JSON type and invalid default"""
        with self.assertRaises(ValueError):
            PluginAppSettingDef(
                name=DEF_NAME,
                scope=APP_SETTING_SCOPE_PROJECT,
                type=APP_SETTING_TYPE_JSON,
                default='{"x": "y"}',
            )

    def test_init_callable_default(self):
        """Test init with callable default"""

        def callable_default(project, user):
            return True

        s_def = PluginAppSettingDef(
            name=DEF_NAME,
            scope=APP_SETTING_SCOPE_PROJECT,
            type=APP_SETTING_TYPE_STRING,
            default=callable_default,
        )
        self.assertIsNotNone(s_def)

    def test_init_callable_options(self):
        """Test init with callable options"""

        def callable_options(project, user):
            return [1, 2]

        s_def = PluginAppSettingDef(
            name=DEF_NAME,
            scope=APP_SETTING_SCOPE_PROJECT,
            type=APP_SETTING_TYPE_INTEGER,
            default=1,
            options=callable_options,
        )
        self.assertIsNotNone(s_def)

    def test_init_default_not_in_options(self):
        """Test init with default value not in options"""
        with self.assertRaises(ValueError):
            PluginAppSettingDef(
                name=DEF_NAME,
                scope=APP_SETTING_SCOPE_PROJECT,
                type=APP_SETTING_TYPE_INTEGER,
                default=0,
                options=[1, 2],
            )

    def test_init_project_user_modifiable_false(self):
        """Test init with PROJECT_USER scope and user_modifiable=False"""
        s_def = PluginAppSettingDef(
            name=DEF_NAME,
            scope=APP_SETTING_SCOPE_PROJECT_USER,
            type=APP_SETTING_TYPE_BOOLEAN,
            default=True,
            user_modifiable=False,
        )
        self.assertIsNotNone(s_def)
        self.assertEqual(s_def.user_modifiable, False)

    def test_init_project_user_modifiable_unset(self):
        """Test init with PROJECT_USER scope and user_modifiable unset"""
        s_def = PluginAppSettingDef(
            name=DEF_NAME,
            scope=APP_SETTING_SCOPE_PROJECT_USER,
            type=APP_SETTING_TYPE_BOOLEAN,
            default=True,
        )
        self.assertIsNotNone(s_def)
        self.assertEqual(s_def.user_modifiable, False)

    def test_init_project_user_modifiable_true(self):
        """Test init with PROJECT_USER scope and user_modifiable=True"""
        with self.assertRaises(ValueError):
            PluginAppSettingDef(
                name=DEF_NAME,
                scope=APP_SETTING_SCOPE_PROJECT_USER,
                type=APP_SETTING_TYPE_BOOLEAN,
                default=True,
                user_modifiable=True,
            )



"""Plugins for the userprofile app"""

from typing import Optional

# Projectroles dependency
from projectroles.models import SODARUser, SODAR_CONSTANTS
from projectroles.plugins import SiteAppPluginPoint, PluginAppSettingDef

from userprofile.urls import urlpatterns


# SODAR constants
APP_SETTING_TYPE_BOOLEAN = SODAR_CONSTANTS['APP_SETTING_TYPE_BOOLEAN']
APP_SETTING_SCOPE_USER = SODAR_CONSTANTS['APP_SETTING_SCOPE_USER']


class SiteAppPlugin(SiteAppPluginPoint):
    """Projectroles plugin for registering the app"""

    #: Name (used as plugin ID)
    name = 'userprofile'

    #: Title (used in templates)
    title = 'User Profile'

    #: UI URLs
    urls = urlpatterns

    #: Iconify icon
    icon = 'mdi:account-details'

    #: Description string
    description = 'Project User Profile'

    #: Entry point URL ID
    entry_point_url_id = 'userprofile:detail'

    #: Required permission for displaying the app
    app_permission = 'userprofile.view_detail'

    #: App setting definitions
    app_settings = [
        PluginAppSettingDef(
            name='enable_project_uuid_copy',
            scope=APP_SETTING_SCOPE_USER,
            type=APP_SETTING_TYPE_BOOLEAN,
            label='Display project UUID copying link',
            description='Display link in project header to copy project UUID '
            'into the clipboard.',
            default=False,
        )
    ]

    def get_messages(self, user: Optional[SODARUser] = None) -> list[dict]:
        """
        Return a list of messages to be shown to users.
        :param user: User object (optional)
        :return: List of dicts or and empty list if no messages
        """
        messages = []
        return messages



"""Django Rest Framework serializers for the filesfolders app"""

# NB: Creating abstract serializers is not easily possible as explained in the
# following StackOverflow post: https://stackoverflow.com/a/33137535

from rest_framework import serializers
from rest_framework.generics import get_object_or_404

# Projectroles dependency
from projectroles.serializers import SODARProjectModelSerializer
from projectroles.utils import build_secret

from filesfolders.models import File, Folder, HyperLink


class FilesfoldersSerializerBase(SODARProjectModelSerializer):
    """Base class for filesfolders serializers"""

    folder = serializers.SlugRelatedField(
        slug_field='sodar_uuid', queryset=Folder.objects.all(), required=False
    )
    owner = serializers.SlugRelatedField(
        slug_field='sodar_uuid', read_only=True
    )

    def _process_validated_data(self, validated_data):
        if 'folder' in self.context['request'].data:
            folder_uuid = self.context['request'].data['folder']
            folder = get_object_or_404(
                Folder.objects.filter(project=self.context['project']),
                sodar_uuid=folder_uuid,
            )
            validated_data['folder'] = folder
        return validated_data

    def update(self, instance, validated_data):
        validated_data = self._process_validated_data(validated_data)
        return super().update(instance, validated_data)

    def create(self, validated_data):
        validated_data['project'] = self.context['project']
        validated_data['owner'] = self.context['request'].user
        validated_data = self._process_validated_data(validated_data)
        return super().create(validated_data)


class FolderSerializer(FilesfoldersSerializerBase):
    """
    Serializer for the Folder model.

    Makes the following fields directly editable: name, flag, description.

    The folder field can be set to a UUID that must refer to a parent folder
    in the parent's folder.

    On creation, the owner is automatically set to the current user that
    cannot be changed later. The project is set to the current project
    on creation and is not updated later.
    """

    class Meta:
        model = Folder
        fields = [
            'name',
            'folder',
            'owner',
            'project',
            'flag',
            'description',
            'date_modified',
            'sodar_uuid',
        ]
        read_only_fields = ['date_modified']


class FileSerializer(FilesfoldersSerializerBase):
    """
    Serializer for the File model.

    Makes the following fields directly editable: name, flag, description.

    The folder field can be set to a UUID that must refer to a parent folder
    in the parent's folder.

    On creation, the owner is automatically set to the current user that
    cannot be changed later. The project is set to the current project
    on creation and is not updated later.

    The secret will be created and updated automatically when the 'public_url'
    flag is changed.
    """

    file = serializers.FileField(write_only=True)

    class Meta:
        model = File
        fields = [
            'name',
            'folder',
            'owner',
            'project',
            'flag',
            'description',
            'date_modified',
            'public_url',
            'secret',
            'file',
            'sodar_uuid',
        ]
        read_only_fields = ['date_modified', 'secret']

    def create(self, validated_data):
        instance = super().create(validated_data)
        instance.secret = build_secret()
        instance.save()
        return instance

    def update(self, instance, validated_data):
        if instance.public_url != bool(
            validated_data.get('public_url', instance.public_url)
        ):
            instance.secret = build_secret()
            instance.save()
        return super().update(instance, validated_data)


class HyperLinkSerializer(FilesfoldersSerializerBase):
    """
    Serializer for the HyperLink model.

    Makes the following fields directly editable: name, flag, description.

    The folder field can be set to a UUID that must refer to a parent folder
    in the parent's folder.

    On creation, the owner is automatically set to the current user that
    cannot be changed later. The project is set to the current project
    on creation and is not updated later.
    """

    class Meta:
        model = HyperLink
        fields = [
            'name',
            'folder',
            'owner',
            'project',
            'flag',
            'description',
            'date_modified',
            'url',
            'sodar_uuid',
        ]
        read_only_fields = ['date_modified']



"""Ajax API views for the timeline app"""

import html

from typing import Any, Optional

from django.http import HttpRequest, HttpResponseForbidden
from django.utils.timezone import localtime
from django.urls import reverse

from rest_framework.response import Response

# Projectroles dependency
from projectroles.views_ajax import (
    SODARBaseProjectAjaxView,
    SODARBasePermissionAjaxView,
    SODARBaseAjaxView,
)

from timeline.models import TimelineEvent, TimelineEventStatus
from timeline.templatetags.timeline_tags import get_status_style


class EventDetailMixin:
    """Mixin for event detail retrieval helpers"""

    @classmethod
    def get_status_extra_url(
        cls, status: TimelineEventStatus, request: HttpRequest
    ) -> Optional[str]:
        """Return URL for extra status data"""
        if status.extra_data == {} or not request.user.has_perm(
            'timeline.view_event_extra_data', status.event.project
        ):
            return None
        return reverse(
            'timeline:ajax_extra_status',
            kwargs={'eventstatus': status.sodar_uuid},
        )

    def get_event_details(
        self, event: TimelineEvent, request: HttpRequest
    ) -> dict:
        """
        Return event details.

        :param event: TimelineEvent object
        :param request: HttpRequest object
        :return: Dict
        """
        ret = {
            'app': event.app,
            'name': event.event_name,
            'user': event.user.username if event.user else 'N/A',
            'timestamp': localtime(event.get_timestamp()).strftime(
                '%Y-%m-%d %H:%M:%S'
            ),
            'status': [],
        }
        status_changes = event.get_status_changes(reverse=True)
        for idx, s in enumerate(status_changes):
            ret['status'].append(
                {
                    'timestamp': localtime(s.timestamp).strftime(
                        '%Y-%m-%d %H:%M:%S'
                    ),
                    'description': s.description,
                    'type': s.status_type,
                    'class': get_status_style(s),
                    'extra_status_link': self.get_status_extra_url(s, request),
                }
            )
        return ret


class EventExtraDataMixin:
    """Mixin for event extra data retrieval helpers"""

    def _json_to_html(self, obj_json: Any) -> str:
        """Return HTML representation of JSON object"""
        str_list = []
        self._html_print_obj(obj_json, str_list, 0)
        str_list.append(
            '<button class="btn btn-secondary sodar-list-btn sodar-copy-btn '
            'sodar-tl-copy-btn" '
            'data-clipboard-target="#data-to-clipboard" '
            'title="Copy to clipboard" data-toggle="tooltip">'
            '<i class="iconify" data-icon="mdi:clipboard-multiple-outline"></i>'
            '</button>'
        )
        return ''.join(str_list)

    def _html_print_obj(self, obj_json: Any, str_list: list, indent: int):
        """Print JSON object to HTML string list"""
        if isinstance(obj_json, dict):
            self._html_print_dict(obj_json, str_list, indent)
        elif isinstance(obj_json, list):
            self._html_print_array(obj_json, str_list, indent)
        elif isinstance(obj_json, str):
            str_list.append('&quot;')
            str_list.append(html.escape(obj_json))
            str_list.append('&quot;')
        elif isinstance(obj_json, int):
            str_list.append(str(obj_json))
        elif isinstance(obj_json, bool):
            str_list.append(str(obj_json))
        elif obj_json is None:
            str_list.append('null')

    def _html_print_dict(self, html_dict: dict, str_list, indent: int):
        """Print JSON dict to HTML string list"""
        str_list.append('<span class="json-open-bracket">{</span><br>')
        str_list.append(
            '<span class="json-collapse-1" style="display: inline;">'
        )
        indent += 1
        for key, value in html_dict.items():
            str_list.append('<span class="json-indent">')
            str_list.append('&nbsp;&nbsp;' * indent)
            str_list.append('</span>')
            str_list.append('<span class="json-property">')
            str_list.append(html.escape(str(key)))
            str_list.append('</span>')
            str_list.append('<span class="json-semi-colon">: </span>')
            str_list.append('<span class="json-value">')
            self._html_print_obj(value, str_list, indent)
            str_list.append('</span>')
            str_list.append('<span class="json-comma">,</span><br>')
        if len(html_dict) > 0:
            del str_list[-1]
            str_list.append('<br>')
        str_list.append('</span>')
        str_list.append('&nbsp;&nbsp;' * (indent - 1))
        str_list.append('<span class="json-close-bracket">}</span>')

    def _html_print_array(self, array: list, str_list: list, indent: int):
        """Print JSON array to HTML string list"""
        str_list.append('<span class="json-open-bracket">[</span><br>')
        str_list.append(
            '<span class="json-collapse-1" style="display: inline;">'
        )
        indent += 1
        for value in array:
            str_list.append('<span class="json-indent">')
            str_list.append('&nbsp;&nbsp;' * indent)
            str_list.append('</span>')
            str_list.append('<span class="json-value">')
            self._html_print_obj(value, str_list, indent)
            str_list.append('</span>')
            str_list.append('<span class="json-comma">,</span><br>')
        if len(array) > 0:
            del str_list[-1]
            str_list.append('<br>')
        str_list.append('</span>')
        str_list.append('&nbsp;&nbsp;' * (indent - 1))
        str_list.append('<span class="json-close-bracket">]</span>')

    def get_event_extra(
        self, event: TimelineEvent, status: Optional[TimelineEventStatus] = None
    ) -> dict:
        """
        Return event extra data.

        :param event: TimelineEvent object
        :param status: TimelineEventStatus object
        :return: JSON-serializable dict
        """
        extra = status.extra_data if status else event.extra_data
        extra_data = self._json_to_html(extra)
        ret = {
            'app': event.app,
            'name': event.event_name,
            'user': event.user.username if event.user else 'N/A',
            'timestamp': localtime(event.get_timestamp()).strftime(
                '%Y-%m-%d %H:%M:%S'
            ),
            'extra': extra_data,
        }
        return ret


class ProjectEventDetailAjaxView(EventDetailMixin, SODARBaseProjectAjaxView):
    """Ajax view for retrieving event details for projects"""

    permission_required = 'timeline.view_timeline'

    def get(self, request, *args, **kwargs):
        event = TimelineEvent.objects.filter(
            sodar_uuid=self.kwargs['timelineevent']
        ).first()
        if event.classified and not request.user.has_perm(
            'timeline.view_classified_event', event.project
        ):
            return HttpResponseForbidden()
        return Response(self.get_event_details(event, request), status=200)


class ProjectEventExtraAjaxView(EventExtraDataMixin, SODARBaseProjectAjaxView):
    """Ajax view for retrieving event extra data for projects"""

    permission_required = 'timeline.view_event_extra_data'

    def get(self, request, *args, **kwargs):
        event = TimelineEvent.objects.filter(
            sodar_uuid=self.kwargs['timelineevent']
        ).first()
        if event.classified and not request.user.has_perm(
            'timeline.view_classified_event', event.project
        ):
            return HttpResponseForbidden()
        return Response(self.get_event_extra(event), status=200)


class SiteEventDetailAjaxView(EventDetailMixin, SODARBasePermissionAjaxView):
    """Ajax view for retrieving event details for site-wide events"""

    permission_required = 'timeline.view_site_timeline'

    def get(self, request, *args, **kwargs):
        event = TimelineEvent.objects.filter(
            sodar_uuid=self.kwargs['timelineevent']
        ).first()
        if event.classified and not request.user.has_perm(
            'timeline.view_classified_site_event'
        ):
            return HttpResponseForbidden()
        return Response(self.get_event_details(event, request), status=200)


class SiteEventExtraAjaxView(EventExtraDataMixin, SODARBasePermissionAjaxView):
    """Ajax view for retrieving event extra data for site-wide events"""

    permission_required = 'timeline.view_event_extra_data'

    def get(self, request, *args, **kwargs):
        event = TimelineEvent.objects.filter(
            sodar_uuid=self.kwargs['timelineevent']
        ).first()
        if event.classified and not request.user.has_perm(
            'timeline.view_classified_site_event'
        ):
            return HttpResponseForbidden()
        return Response(self.get_event_extra(event), status=200)


class EventStatusExtraAjaxView(EventExtraDataMixin, SODARBaseAjaxView):
    """Ajax view for retrieving event status extra data for events"""

    def get(self, request, *args, **kwargs):
        status = TimelineEventStatus.objects.filter(
            sodar_uuid=self.kwargs['eventstatus']
        ).first()
        event = status.event
        if event.classified and event.project:
            perm = 'view_classified_event'
        elif event.classified:
            perm = 'view_classified_site_event'
        else:
            perm = 'view_event_extra_data'
        project = getattr(event, 'project', None)
        if not request.user.has_perm('timeline.' + perm, project):
            return HttpResponseForbidden()
        return Response(self.get_event_extra(status.event, status), status=200)



"""Tests for models in the adminalerts app"""

from django.forms.models import model_to_dict
from django.utils import timezone

from test_plus.test import TestCase

# Projectroles dependency
from projectroles.models import SODARUser

from adminalerts.models import AdminAlert


class AdminAlertMixin:
    """Helper mixin for AdminAlert creation"""

    @classmethod
    def make_alert(
        cls,
        message: str,
        user: SODARUser,
        description: str,
        active: bool = True,
        require_auth: bool = True,
        date_expire_days: int = 1,
    ) -> AdminAlert:
        """Create AdminAlert object"""
        values = {
            'message': message,
            'user': user,
            'description': description,
            'date_expire': timezone.now()
            + timezone.timedelta(days=date_expire_days),
            'active': active,
            'require_auth': require_auth,
        }
        return AdminAlert.objects.create(**values)


class TestAdminAlert(AdminAlertMixin, TestCase):
    """Tests for AdminAlert model"""

    def setUp(self):
        # Create superuser
        self.superuser = self.make_user('superuser')
        self.superuser.is_superuser = True
        # Create alert
        self.alert = self.make_alert(
            message='alert',
            user=self.superuser,
            description='description',
            active=True,
            require_auth=True,
        )

    def test_initialization(self):
        expected = {
            'id': self.alert.pk,
            'message': 'alert',
            'user': self.superuser.pk,
            'date_expire': self.alert.date_expire,
            'active': True,
            'require_auth': True,
            'sodar_uuid': self.alert.sodar_uuid,
        }
        model_dict = model_to_dict(self.alert)
        # HACK: Can't compare markupfields like this. Better solution?
        model_dict.pop('description', None)
        self.assertEqual(model_dict, expected)

    def test__str__(self):
        expected = 'alert [ACTIVE]'
        self.assertEqual(str(self.alert), expected)

    def test__repr__(self):
        expected = "AdminAlert('alert', 'superuser', True)"
        self.assertEqual(repr(self.alert), expected)



"""API tests for the timeline app"""

from django.contrib.auth.models import AnonymousUser
from django.http import HttpRequest
from django.forms.models import model_to_dict
from django.test import RequestFactory
from django.urls import reverse

# Projectroles dependency
from projectroles.models import Project, SODARUser, SODAR_CONSTANTS
from projectroles.plugins import PluginAPI
from projectroles.tests.test_models import (
    RemoteSiteMixin,
    REMOTE_SITE_NAME,
    REMOTE_SITE_URL,
)

# Filesfolders dependency
from filesfolders.tests.test_models import FolderMixin

from timeline.models import (
    TimelineEvent,
    TimelineEventStatus,
    TimelineEventObjectRef,
    DEFAULT_MESSAGES,
)
from timeline.templatetags import timeline_tags as tags
from timeline.tests.test_models import (
    TimelineEventTestBase,
    TimelineEventMixin,
    TimelineEventStatusMixin,
    EXTRA_DATA,
)


plugin_api = PluginAPI()


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

# Local constants
APP_NAME_PR = 'projectroles'
APP_NAME_FF = 'filesfolders'


class TimelineAPIMixin:
    """Helpers for timeline API use"""

    @classmethod
    def get_request(cls, user: SODARUser, project: Project) -> HttpRequest:
        """Return mock request"""
        request = RequestFactory().get(
            'timeline:list', kwargs={'project': project}
        )
        request.user = user
        return request


class TestTimelineAPI(
    TimelineAPIMixin,
    TimelineEventMixin,
    TimelineEventStatusMixin,
    RemoteSiteMixin,
    FolderMixin,
    TimelineEventTestBase,
):
    def setUp(self):
        super().setUp()
        self.timeline = plugin_api.get_backend_api('timeline_backend')
        self.superuser = self.make_user('superuser')
        self.superuser.is_superuser = True
        self.superuser.save()

    def test_add_event(self):
        """Test adding an event"""
        self.assertEqual(TimelineEvent.objects.all().count(), 0)
        self.assertEqual(TimelineEventStatus.objects.all().count(), 0)

        event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.user_owner,
            event_name='test_event',
            description='description',
            extra_data=EXTRA_DATA,
        )

        self.assertEqual(TimelineEvent.objects.all().count(), 1)
        self.assertEqual(TimelineEventStatus.objects.all().count(), 1)  # Init

        expected = {
            'id': event.pk,
            'project': self.project.pk,
            'app': APP_NAME_PR,
            'plugin': None,
            'user': self.user_owner.pk,
            'event_name': 'test_event',
            'description': 'description',
            'classified': False,
            'extra_data': EXTRA_DATA,
            'sodar_uuid': event.sodar_uuid,
        }
        self.assertEqual(model_to_dict(event), expected)

        status = event.get_status()
        expected_status = {
            'id': status.pk,
            'event': event.pk,
            'status_type': self.timeline.TL_STATUS_INIT,
            'description': DEFAULT_MESSAGES[self.timeline.TL_STATUS_INIT],
            'extra_data': {},
            'sodar_uuid': status.sodar_uuid,
        }
        self.assertEqual(model_to_dict(status), expected_status)

    def test_add_event_with_status(self):
        """Test adding an event with status"""
        self.assertEqual(TimelineEvent.objects.all().count(), 0)
        self.assertEqual(TimelineEventStatus.objects.all().count(), 0)

        event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.user_owner,
            event_name='test_event',
            description='description',
            extra_data=EXTRA_DATA,
            status_type=self.timeline.TL_STATUS_OK,
            status_desc=self.timeline.TL_STATUS_OK,
            status_extra_data={},
        )

        status = event.get_status()
        self.assertEqual(TimelineEvent.objects.all().count(), 1)
        self.assertEqual(TimelineEventStatus.objects.all().count(), 2)

        expected_event = {
            'id': event.pk,
            'project': self.project.pk,
            'app': APP_NAME_PR,
            'plugin': None,
            'user': self.user_owner.pk,
            'event_name': 'test_event',
            'description': 'description',
            'classified': False,
            'extra_data': EXTRA_DATA,
            'sodar_uuid': event.sodar_uuid,
        }
        self.assertEqual(model_to_dict(event), expected_event)

        expected_status = {
            'id': status.pk,
            'event': event.pk,
            'status_type': self.timeline.TL_STATUS_OK,
            'description': self.timeline.TL_STATUS_OK,
            'extra_data': {},
            'sodar_uuid': status.sodar_uuid,
        }
        self.assertEqual(model_to_dict(status), expected_status)

    def test_add_event_custom_init(self):
        """Test adding an event with custom INIT status"""
        custom_init_desc = 'Custom init'
        self.assertEqual(TimelineEvent.objects.all().count(), 0)
        self.assertEqual(TimelineEventStatus.objects.all().count(), 0)

        event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.user_owner,
            event_name='test_event',
            description='description',
            extra_data=EXTRA_DATA,
            status_type=self.timeline.TL_STATUS_INIT,
            status_desc=custom_init_desc,
        )

        self.assertEqual(TimelineEvent.objects.all().count(), 1)
        self.assertEqual(TimelineEventStatus.objects.all().count(), 1)  # Init

        status = event.get_status()
        expected_status = {
            'id': status.pk,
            'event': event.pk,
            'status_type': self.timeline.TL_STATUS_INIT,
            'description': custom_init_desc,
            'extra_data': {},
            'sodar_uuid': status.sodar_uuid,
        }
        self.assertEqual(model_to_dict(status), expected_status)

    def test_add_event_no_user(self):
        """Test adding an event with no user"""
        self.assertEqual(TimelineEvent.objects.all().count(), 0)
        self.assertEqual(TimelineEventStatus.objects.all().count(), 0)

        event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=None,
            event_name='test_event',
            description='description',
            extra_data=EXTRA_DATA,
        )

        self.assertEqual(TimelineEvent.objects.all().count(), 1)
        self.assertEqual(TimelineEventStatus.objects.all().count(), 1)

        expected = {
            'id': event.pk,
            'project': self.project.pk,
            'app': APP_NAME_PR,
            'plugin': None,
            'user': None,
            'event_name': 'test_event',
            'description': 'description',
            'classified': False,
            'extra_data': EXTRA_DATA,
            'sodar_uuid': event.sodar_uuid,
        }
        self.assertEqual(model_to_dict(event), expected)

    def test_add_event_anon_user(self):
        """Test adding an event with AnonymousUser"""
        self.assertEqual(TimelineEvent.objects.all().count(), 0)
        self.assertEqual(TimelineEventStatus.objects.all().count(), 0)

        event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=AnonymousUser(),
            event_name='test_event',
            description='description',
            extra_data=EXTRA_DATA,
        )

        self.assertEqual(TimelineEvent.objects.all().count(), 1)
        self.assertEqual(TimelineEventStatus.objects.all().count(), 1)
        self.assertIsNone(event.user)

    def test_add_event_invalid_app(self):
        """Test adding an event with an invalid app name (should fail)"""
        self.assertEqual(TimelineEvent.objects.all().count(), 0)
        self.assertEqual(TimelineEventStatus.objects.all().count(), 0)

        with self.assertRaises(ValueError):
            self.timeline.add_event(
                project=self.project,
                app_name='NON-EXISTING APP NAME',
                user=self.user_owner,
                event_name='test_event',
                description='description',
                extra_data=EXTRA_DATA,
            )

        self.assertEqual(TimelineEvent.objects.all().count(), 0)
        self.assertEqual(TimelineEventStatus.objects.all().count(), 0)

    def test_add_event_invalid_status(self):
        """Test adding an event with an invalid status type"""
        self.assertEqual(TimelineEvent.objects.all().count(), 0)
        self.assertEqual(TimelineEventStatus.objects.all().count(), 0)

        with self.assertRaises(ValueError):
            self.timeline.add_event(
                project=self.project,
                app_name=APP_NAME_PR,
                user=self.user_owner,
                event_name='test_event',
                description='description',
                status_type='NON-EXISTING STATUS TYPE',
                extra_data=EXTRA_DATA,
            )

        self.assertEqual(TimelineEvent.objects.all().count(), 0)
        self.assertEqual(TimelineEventStatus.objects.all().count(), 0)

    def test_add_object(self):
        """Test adding an object to an event"""
        self.assertEqual(TimelineEventObjectRef.objects.all().count(), 0)

        event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.user_owner,
            event_name='test_event',
            description='event with {obj}',
            extra_data=EXTRA_DATA,
        )
        temp_obj = self.project.get_owner()
        ref = event.add_object(
            obj=temp_obj,
            label='obj',
            name='assignment',
            extra_data=EXTRA_DATA,
        )

        self.assertEqual(TimelineEventObjectRef.objects.all().count(), 1)
        expected = {
            'id': ref.pk,
            'event': event.pk,
            'label': 'obj',
            'name': 'assignment',
            'object_model': temp_obj.__class__.__name__,
            'object_uuid': temp_obj.sodar_uuid,
            'extra_data': EXTRA_DATA,
            'sodar_uuid': ref.sodar_uuid,
        }
        self.assertEqual(model_to_dict(ref), expected)

    def test_get_project_events(self):
        """Test get_project_events()"""
        event_normal = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.user_owner,
            event_name='test_event',
            description='description',
            extra_data=EXTRA_DATA,
        )
        event_classified = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.user_owner,
            event_name='test_event',
            description='description',
            classified=True,
            extra_data=EXTRA_DATA,
        )

        events = self.timeline.get_project_events(
            self.project, classified=False
        )
        self.assertEqual(events.count(), 1)
        self.assertEqual(events[0], event_normal)

        # Test with classified
        events = self.timeline.get_project_events(self.project, classified=True)
        self.assertEqual(events.count(), 2)
        self.assertIn(event_classified, events)

    def test_get_object_url(self):
        """Test get_object_url()"""
        expected_url = reverse(
            'timeline:list_object',
            kwargs={
                'project': self.project.sodar_uuid,
                'object_model': self.user_owner.__class__.__name__,
                'object_uuid': self.user_owner.sodar_uuid,
            },
        )
        url = self.timeline.get_object_url(self.user_owner, self.project)
        self.assertEqual(expected_url, url)

    def test_get_object_url_no_project(self):
        """Test get_object_url() without project"""
        expected_url = reverse(
            'timeline:list_object_site',
            kwargs={
                'object_model': self.user_owner.__class__.__name__,
                'object_uuid': self.user_owner.sodar_uuid,
            },
        )
        url = self.timeline.get_object_url(self.user_owner, None)
        self.assertEqual(expected_url, url)

    def test_get_object_link(self):
        """Test get_object_link()"""
        expected_url = reverse(
            'timeline:list_object',
            kwargs={
                'project': self.project.sodar_uuid,
                'object_model': self.user_owner.__class__.__name__,
                'object_uuid': self.user_owner.sodar_uuid,
            },
        )
        link = self.timeline.get_object_link(self.user_owner, self.project)
        self.assertIn(expected_url, link)

    def test_get_object_link_no_project(self):
        """Test get_object_link() without project"""
        expected_url = reverse(
            'timeline:list_object_site',
            kwargs={
                'object_model': self.user_owner.__class__.__name__,
                'object_uuid': self.user_owner.sodar_uuid,
            },
        )
        link = self.timeline.get_object_link(self.user_owner, None)
        self.assertIn(expected_url, link)

    def test_get_event_description(self):
        """Test getting event description"""
        event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.user_owner,
            event_name='test_event',
            description='description',
            extra_data=EXTRA_DATA,
        )
        self.assertEqual(
            self.timeline.get_event_description(event), event.description
        )

    def test_get_event_description_user(self):
        """Test getting event description with User object"""
        event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.user_owner,
            event_name='test_event',
            description='event with {obj}',
        )
        event.add_object(
            obj=self.user_owner,
            label='obj',
            name=self.user_owner.username,
        )
        desc = self.timeline.get_event_description(event)
        self.assertIn(self.user_owner.username, desc)
        self.assertIn('sodar-tl-object-link', desc)

    def test_get_event_description_project(self):
        """Test getting event description with Project object"""
        event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.user_owner,
            event_name='test_event',
            description='event with {obj}',
        )
        event.add_object(
            obj=self.project,
            label='obj',
            name=self.project.title,
        )
        desc = self.timeline.get_event_description(
            event, request=self.get_request(self.user_owner, self.project)
        )
        self.assertIn(self.project.title, desc)
        self.assertIn('sodar-tl-project-link', desc)

    def test_get_event_description_remote_site(self):
        """Test getting event description with RemoteSite object"""
        event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.user_owner,
            event_name='test_event',
            description='event with {obj}',
        )
        site = self.make_site(name=REMOTE_SITE_NAME, url=REMOTE_SITE_URL)
        event.add_object(
            obj=site,
            label='obj',
            name=site.name,
        )
        desc = self.timeline.get_event_description(
            event, request=self.get_request(self.superuser, self.project)
        )
        self.assertIn(site.name, desc)
        self.assertIn('sodar-tl-object-link', desc)

    def test_get_event_description_app(self):
        """Test getting event description with app plugin"""
        event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_FF,
            user=self.user_owner,
            event_name='test_event',
            description='event with {obj}',
        )
        folder = self.make_folder(
            name='folder',
            project=self.project,
            folder=None,
            description='',
            owner=self.user_owner,
        )
        event.add_object(
            obj=folder,
            label='obj',
            name=folder.name,
        )
        desc = self.timeline.get_event_description(
            event, request=self.get_request(self.superuser, self.project)
        )
        self.assertIn(folder.name, desc)
        self.assertIn('sodar-tl-object-link', desc)

    def test_get_event_description_lookup(self):
        """Test getting event description with app plugin and lookup dict"""
        event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_FF,
            user=self.user_owner,
            event_name='test_event',
            description='event with {obj}',
        )
        folder = self.make_folder(
            name='folder',
            project=self.project,
            folder=None,
            description='',
            owner=self.user_owner,
        )
        event.add_object(
            obj=folder,
            label='obj',
            name=folder.name,
        )
        desc = self.timeline.get_event_description(
            event,
            plugin_lookup=tags.get_plugin_lookup(),
            request=self.get_request(self.superuser, self.project),
        )
        self.assertIn(folder.name, desc)
        self.assertIn('sodar-tl-object-link', desc)

    def test_get_event_description_invalid_app(self):
        """Test getting event description with non-existing app plugin"""
        event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_PR,
            user=self.user_owner,
            event_name='test_event',
            description='event with {obj}',
        )
        event.app = 'NOT_AN_ACTUAL_APP'
        event.save()
        event.add_object(
            obj=self.project,
            label='obj',
            name=self.project.title,
        )
        desc = self.timeline.get_event_description(
            event, request=self.get_request(self.superuser, self.project)
        )
        self.assertNotIn(self.project.title, desc)
        self.assertIn('sodar-tl-plugin-error', desc)

    def test_get_event_description_specified_plugin(self):
        """Test getting event description with specified app plugin"""
        event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_FF,
            plugin_name=APP_NAME_FF,
            user=self.user_owner,
            event_name='test_event',
            description='event with {obj}',
        )
        folder = self.make_folder(
            name='folder',
            project=self.project,
            folder=None,
            description='',
            owner=self.user_owner,
        )
        event.add_object(
            obj=folder,
            label='obj',
            name=folder.name,
        )
        desc = self.timeline.get_event_description(
            event, request=self.get_request(self.superuser, self.project)
        )
        self.assertIn(folder.name, desc)
        self.assertIn('sodar-tl-object-link', desc)

    def test_get_event_description_invalid_plugin(self):
        """Test description with an invalid specified app plugin"""
        event = self.timeline.add_event(
            project=self.project,
            app_name=APP_NAME_FF,
            plugin_name='NOT_A_REAL_PLUGIN',
            user=self.user_owner,
            event_name='test_event',
            description='event with {obj}',
        )
        folder = self.make_folder(
            name='folder',
            project=self.project,
            folder=None,
            description='',
            owner=self.user_owner,
        )
        event.add_object(
            obj=folder,
            label='obj',
            name=folder.name,
        )
        desc = self.timeline.get_event_description(
            event, request=self.get_request(self.superuser, self.project)
        )
        self.assertNotIn(folder.name, desc)
        self.assertIn('sodar-tl-plugin-error', desc)



"""URLs for the timeline app"""

from django.urls import path

from timeline import views, views_ajax, views_api


app_name = 'timeline'

# UI views for project plugin
urls_ui_project = [
    path(
        route='<uuid:project>',
        view=views.ProjectTimelineView.as_view(),
        name='list_project',
    ),
    path(
        route='<uuid:project>/<str:object_model>/<uuid:object_uuid>',
        view=views.ProjectObjectTimelineView.as_view(),
        name='list_object',
    ),
]

# UI views for site app plugin
urls_ui_site = [
    path(route='site', view=views.SiteTimelineView.as_view(), name='list_site'),
    path(
        route='site/<str:object_model>/<uuid:object_uuid>',
        view=views.SiteObjectTimelineView.as_view(),
        name='list_object_site',
    ),
]

# UI views for admin site app plugin
urls_ui_admin = [
    path(
        route='site/all',
        view=views.AdminTimelineView.as_view(),
        name='list_admin',
    ),
]

# Ajax API views
urls_ajax = [
    path(
        route='ajax/detail/<uuid:timelineevent>',
        view=views_ajax.ProjectEventDetailAjaxView.as_view(),
        name='ajax_detail_project',
    ),
    path(
        route='ajax/detail/site/<uuid:timelineevent>',
        view=views_ajax.SiteEventDetailAjaxView.as_view(),
        name='ajax_detail_site',
    ),
    path(
        route='ajax/extra/<uuid:timelineevent>',
        view=views_ajax.ProjectEventExtraAjaxView.as_view(),
        name='ajax_extra_project',
    ),
    path(
        route='ajax/extra/site/<uuid:timelineevent>',
        view=views_ajax.SiteEventExtraAjaxView.as_view(),
        name='ajax_extra_site',
    ),
    path(
        route='ajax/extra/status/<uuid:eventstatus>',
        view=views_ajax.EventStatusExtraAjaxView.as_view(),
        name='ajax_extra_status',
    ),
]

# REST API views
urls_api = [
    path(
        route='api/list/<uuid:project>',
        view=views_api.ProjectTimelineEventListAPIView.as_view(),
        name='api_list_project',
    ),
    path(
        route='api/list/site',
        view=views_api.SiteTimelineEventListAPIView.as_view(),
        name='api_list_site',
    ),
    path(
        route='api/retrieve/<uuid:timelineevent>',
        view=views_api.TimelineEventRetrieveAPIView.as_view(),
        name='api_retrieve',
    ),
]

urlpatterns = (
    urls_ui_project + urls_ui_site + urls_ui_admin + urls_ajax + urls_api
)



"""
Cleanappsettings management command for cleaning up undefined or orphaned app
settings.
"""

from django.core.management.base import BaseCommand

from projectroles.app_settings import AppSettingAPI
from projectroles.management.logging import ManagementCommandLogger
from projectroles.models import AppSetting, SODAR_CONSTANTS


app_settings = AppSettingAPI()
logger = ManagementCommandLogger(__name__)

# SODAR constants
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']
APP_SETTING_SCOPE_USER = SODAR_CONSTANTS['APP_SETTING_SCOPE_USER']

# Local constants
APP_NAME = 'projectroles'
START_MSG = 'Querying database for undefined app settings..'
CHECK_MODE_MSG = 'Check mode enabled, database will not be altered'
END_MSG = 'OK'
DEF_NOT_FOUND_MSG = 'definition not found'
ALLOWED_TYPES_MSG = 'does not match allowed types'
LOG_NONE_LABEL = 'None'
PREFIX_MSG = '"{s_name}" (project={project}; user={user}): '
CHECK_PREFIX_MSG = 'Found ' + PREFIX_MSG
DELETE_PREFIX_MSG = 'Deleting ' + PREFIX_MSG
DELETE_PROJECT_TYPE_MSG = 'project type "{}" {}: {}'
DELETE_SCOPE_MSG = 'user has no role in project'


class Command(BaseCommand):
    """Command for cleaning up undefined or orphaned app settings"""

    help = (
        'Cleans up undefined or otherwise orphaned app settings from the '
        'database.'
    )

    @classmethod
    def _get_log_msg(cls, s: AppSetting, msg: str, check: bool) -> str:
        """Return delete/check message for logging"""
        prefix_msg = CHECK_PREFIX_MSG if check else DELETE_PREFIX_MSG
        s_name = '.'.join(
            [
                'settings',
                APP_NAME if s.app_plugin is None else s.app_plugin.name,
                s.name,
            ]
        )
        p_str = f'"{s.project.title}"' if s.project else LOG_NONE_LABEL
        u_str = f'"{s.user.username}"' if s.user else LOG_NONE_LABEL
        return prefix_msg.format(s_name=s_name, project=p_str, user=u_str) + msg

    def add_arguments(self, parser):
        parser.add_argument(
            '-c',
            '--check',
            dest='check',
            required=False,
            default=False,
            action='store_true',
            help='Log settings to be cleaned up without altering the database',
        )
        parser.add_argument(
            '-s',
            '--superuser',
            dest='superuser',
            required=False,
            default=False,
            action='store_true',
            help='Clear PROJECT_USER settings from superusers with no roles',
        )

    def handle(self, *args, **options):
        check = options.get('check', False)
        if check:
            logger.info(CHECK_MODE_MSG)
        logger.info(START_MSG)
        s_def = None
        p_types = []
        db_settings = AppSetting.objects.all()
        for s in db_settings:
            # Undefined settings
            def_kwargs = {'name': s.name}
            if s.app_plugin:
                def_kwargs['plugin'] = s.app_plugin.get_plugin()
            else:
                def_kwargs['plugin_name'] = APP_NAME
            try:
                s_def = app_settings.get_definition(**def_kwargs)
                # Get allowed project types (if unset, default is PROJECT only)
                if s_def.scope != APP_SETTING_SCOPE_USER:
                    p_types = s_def.project_types
            except ValueError:
                logger.info(self._get_log_msg(s, DEF_NOT_FOUND_MSG, check))
                if not check:
                    s.delete()
                    continue
            # Invalid scope
            if s_def and s.project and s.project.type not in p_types:
                msg = DELETE_PROJECT_TYPE_MSG.format(
                    s.project.type, ALLOWED_TYPES_MSG, ', '.join(p_types)
                )
                logger.info(self._get_log_msg(s, msg, check))
                if not check:
                    s.delete()
                    continue
            # No user role for PROJECT_USER scope setting
            if (
                s.project
                and s.user
                and (not s.user.is_superuser or options.get('superuser', False))
                and not s.project.get_role(s.user)
            ):
                logger.info(self._get_log_msg(s, DELETE_SCOPE_MSG, check))
                if not check:
                    s.delete()
        logger.info(END_MSG)



"""General utility methods for projectroles and SODAR Core"""

import logging
import random
import string

from typing import Optional

from django.conf import settings

from projectroles.models import SODAR_CONSTANTS


logger = logging.getLogger(__name__)


def get_display_name(
    key: str, title: bool = False, count: int = 1, plural: bool = False
) -> str:
    """
    Return display name from SODAR_CONSTANTS.

    :param key: Key in SODAR_CONSTANTS['DISPLAY_NAMES'] to return (string)
    :param title: Return name in title case if true (boolean, optional)
    :param count: Item count for returning plural form, overrides plural=False
                  if not 1 (int, optional)
    :param plural: Return plural form if True, overrides count != 1 if True
                   (boolean, optional)
    :return: String
    """
    ret = SODAR_CONSTANTS['DISPLAY_NAMES'][key][
        'plural' if count != 1 or plural else 'default'
    ]
    return ret.lower() if not title else ret.title()


def build_secret(length: Optional[int] = None) -> str:
    """
    Return secret string for e.g. public URLs.

    :param length: Length of string, use None for default (integer or None)
    :return: Randomized secret (string)
    """
    if not length:
        length = getattr(settings, 'PROJECTROLES_SECRET_LENGTH', 32)
    length = int(length) if int(length) <= 255 else 255
    return ''.join(
        random.SystemRandom().choice(string.ascii_lowercase + string.digits)
        for _ in range(length)
    )


def get_app_names() -> list[str]:
    """Return list of names for locally installed non-django apps"""
    ret = []
    for a in settings.INSTALLED_APPS:
        s = a.split('.')
        if s[0] not in ['django', settings.SITE_PACKAGE]:
            if len(s) > 1 and 'apps' in s:
                ret.append('.'.join(s[0 : s.index('apps')]))
            else:
                ret.append(s[0])
    return sorted(ret)



"""Models for the timeline app"""

import logging
import uuid

from datetime import datetime
from typing import Any, Optional, Union

from django.conf import settings
from django.db import models
from django.db.models import Max, Q, QuerySet

# Projectroles dependency
from projectroles.models import Project


logger = logging.getLogger(__name__)
# Access Django user model
AUTH_USER_MODEL = getattr(settings, 'AUTH_USER_MODEL', 'auth.User')

# Local constants
TL_STATUS_OK = 'OK'
TL_STATUS_INIT = 'INIT'
TL_STATUS_SUBMIT = 'SUBMIT'
TL_STATUS_FAILED = 'FAILED'
TL_STATUS_INFO = 'INFO'
TL_STATUS_CANCEL = 'CANCEL'
EVENT_STATUS_TYPES = [
    TL_STATUS_OK,
    TL_STATUS_INIT,
    TL_STATUS_SUBMIT,
    TL_STATUS_FAILED,
    TL_STATUS_INFO,
    TL_STATUS_CANCEL,
]
DEFAULT_MESSAGES = {
    TL_STATUS_OK: 'All OK',
    TL_STATUS_INIT: 'Event initialized',
    TL_STATUS_SUBMIT: 'Job submitted asynchronously',
    TL_STATUS_FAILED: 'Failed (unknown problem)',
    TL_STATUS_INFO: 'Info level action',
    TL_STATUS_CANCEL: 'Action cancelled',
}
OBJ_REF_UNNAMED = '(unnamed)'


class TimelineEventManager(models.Manager):
    """Manager for custom table-level TimelineEvent queries"""

    def get_object_events(
        self,
        project: Optional[Project],
        object_model: str,
        object_uuid: Union[str, uuid.UUID],
        order_by: str = '-pk',
    ) -> QuerySet:
        """
        Return events which are linked to an object reference.

        :param project: Project object or None
        :param object_model: Object model (string)
        :param object_uuid: sodar_uuid of the original object
        :param order_by: Ordering (default = pk descending)
        :return: QuerySet
        """
        return TimelineEvent.objects.filter(
            project=project,
            event_objects__object_model=object_model,
            event_objects__object_uuid=object_uuid,
        ).order_by(order_by)

    def find(
        self, search_terms: list[str], keywords: Optional[dict] = None
    ) -> QuerySet:
        """
        Return events matching the query.

        :param search_terms: Search terms (list of strings)
        :param keywords: Optional search keywords as key/value pairs (dict)
        :return: QuerySet of TimelineEvent objects
        """
        search_limit = getattr(settings, 'TIMELINE_SEARCH_LIMIT', 250)
        term_query = Q()
        for t in search_terms:
            term_query.add(Q(event_name__icontains=t), Q.OR)
            term_query.add(Q(event_name__icontains=t.replace(' ', '_')), Q.OR)
            term_query.add(Q(description__icontains=t), Q.OR)
            term_query.add(Q(event_objects__name__icontains=t), Q.OR)
            term_query.add(Q(user__name__icontains=t), Q.OR)
            term_query.add(Q(user__username__icontains=t), Q.OR)
        items = (
            super()
            .get_queryset()
            .filter(term_query)
            .annotate(timestamp=Max('status_changes__timestamp'))
            .order_by('-timestamp')
        )
        return items[:search_limit]


class TimelineEvent(models.Model):
    """
    Class representing a Project event. Can also be a site-wide event not linked
    to a specific project.
    """

    #: Project to which the event belongs
    project = models.ForeignKey(
        Project,
        related_name='events',
        help_text='Project to which the event belongs (null for no project)',
        on_delete=models.CASCADE,
        null=True,
    )

    #: App from which the event was triggered
    app = models.CharField(
        max_length=255, help_text='App from which the event was triggered'
    )

    #: Plugin to which the event is related (optional)
    plugin = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        help_text='Plugin to which the event is related (optional, if unset '
        'plugin with the name of the app is assumed)',
    )

    #: User who initiated the event (optional)
    user = models.ForeignKey(
        AUTH_USER_MODEL,
        null=True,
        help_text='User who initiated the event (optional)',
        on_delete=models.CASCADE,
    )

    #: Event ID string
    event_name = models.CharField(max_length=255, help_text='Event ID string')

    #: Description of status change (may include {object_name} references)
    description = models.TextField(
        help_text='Description of status change '
        '(may include {object label} references)'
    )

    #: Additional event data as JSON
    extra_data = models.JSONField(
        default=dict, help_text='Additional event data as JSON'
    )

    #: Event is classified (only viewable by user levels specified in rules)
    classified = models.BooleanField(
        default=False,
        help_text='Event is classified (only viewable by user levels '
        'specified in rules)',
    )

    #: UUID for the event
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4, unique=True, help_text='Event SODAR UUID'
    )

    # Set manager for custom queries
    objects = TimelineEventManager()

    def __str__(self):
        return '{}{}{}'.format(
            (self.project.title + ': ') if self.project else '',
            self.event_name,
            ('/' + self.user.username) if self.user else '',
        )

    def __repr__(self):
        return 'TimelineEvent({})'.format(
            ', '.join(repr(v) for v in self.get_repr_values())
        )

    def get_repr_values(self) -> list[str]:
        return [
            self.project.title if self.project else 'N/A',
            self.event_name,
            self.user.username if self.user else 'N/A',
        ]

    def get_status(self) -> Optional['TimelineEventStatus']:
        """Return the current event status"""
        return self.status_changes.order_by('-timestamp').first()

    def get_timestamp(self) -> datetime:
        """Return the timestamp of current status"""
        return self.status_changes.order_by('-timestamp').first().timestamp

    def get_status_changes(self, reverse: bool = False) -> QuerySet:
        """Return all status changes for the event"""
        return self.status_changes.order_by(
            '{}pk'.format('-' if reverse else '')
        )

    def get_project(self) -> Project:
        """Return the project for the event"""
        return self.project

    def add_object(
        self, obj: Any, label: str, name: str, extra_data: Optional[dict] = None
    ) -> 'TimelineEventObjectRef':
        """
        Add object reference to an event.

        :param obj: Django object to which we want to refer
        :param label: Label for the object in the event description (string)
        :param name: Name or title of the object (string)
        :param extra_data: Additional data related to object (dict, optional)
        :return: TimelineEventObjectRef object
        """
        ref = TimelineEventObjectRef()
        ref.event = self
        ref.label = label
        if not name:
            logger.warning(
                'Adding object reference with no name: "{}" ({})'.format(
                    obj, getattr(obj, 'sodar_uuid')
                )
            )
            name = OBJ_REF_UNNAMED
        ref.name = name
        ref.object_model = obj.__class__.__name__
        ref.object_uuid = obj.sodar_uuid
        if extra_data:
            ref.extra_data = extra_data
        ref.save()
        return ref

    def set_status(
        self,
        status_type: str,
        status_desc: Optional[str] = None,
        extra_data: Optional[dict] = None,
    ) -> 'TimelineEventStatus':
        """
        Set event status.

        :param status_type: Status type string (see EVENT_STATUS_TYPES)
        :param status_desc: Description string (optional)
        :param extra_data: Extra data for the status (dict, optional)
        :return: TimelineEventStatus object
        :raise: TypeError if status_type is invalid
        """
        if status_type not in EVENT_STATUS_TYPES:
            raise TypeError(
                'Invalid status type (accepted values: {})'.format(
                    ', '.join(v for v in EVENT_STATUS_TYPES)
                )
            )
        status = TimelineEventStatus()
        status.event = self
        status.status_type = status_type
        status.description = (
            status_desc if status_desc else DEFAULT_MESSAGES[status_type]
        )
        if extra_data:
            status.extra_data = extra_data
        status.save()
        return status


class TimelineEventObjectRef(models.Model):
    """
    Class representing a reference to an object (existing or removed)
    related to a timeline event status.
    """

    #: Event to which the object belongs
    event = models.ForeignKey(
        TimelineEvent,
        related_name='event_objects',
        help_text='Event to which the object belongs',
        on_delete=models.CASCADE,
    )

    #: Label for the object related to the event
    label = models.CharField(
        max_length=255,
        null=False,
        blank=False,
        help_text='Label for the object related to the event',
    )

    #: Name or title of the object
    name = models.CharField(
        max_length=255,
        null=False,
        blank=False,
        help_text='Name or title of the object',
    )

    #: Object model as string
    object_model = models.CharField(
        max_length=255,
        null=False,
        blank=False,
        help_text='Object model as string',
    )

    #: Object SODAR UUID
    object_uuid = models.UUIDField(
        null=True, blank=True, unique=False, help_text='Object SODAR UUID'
    )

    #: Additional data related to the object as JSON
    extra_data = models.JSONField(
        default=dict, help_text='Additional data related to the object as JSON'
    )

    #: UUID for this object reference
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4, unique=True, help_text='Object reference SODAR UUID'
    )

    def __str__(self):
        return '{} ({})'.format(
            self.event.__str__(),
            self.name,
        )

    def __repr__(self):
        values = self.event.get_repr_values() + [self.name]
        return 'TimelineEventObjectRef({})'.format(
            ', '.join(repr(v) for v in values)
        )

    def get_project(self) -> Project:
        """Return the project for the event"""
        return self.event.project


class TimelineEventStatus(models.Model):
    """Class representing a timeline event status"""

    #: Event to which the status change belongs
    event = models.ForeignKey(
        TimelineEvent,
        related_name='status_changes',
        help_text='Event to which the status change belongs',
        on_delete=models.CASCADE,
    )

    #: DateTime of the status change
    timestamp = models.DateTimeField(
        auto_now_add=True, help_text='DateTime of the status change'
    )

    #: Type of the status change
    status_type = models.CharField(
        max_length=64,
        null=False,
        blank=False,
        help_text='Type of the status change',
    )

    #: Description of status change (optional)
    description = models.TextField(
        blank=True, help_text='Description of status change (optional)'
    )

    #: Additional status data as JSON
    extra_data = models.JSONField(
        default=dict, help_text='Additional status data as JSON'
    )

    #: UUID for the status
    sodar_uuid = models.UUIDField(
        default=uuid.uuid4, unique=True, help_text='Status SODAR UUID'
    )

    def __str__(self):
        return '{} ({})'.format(
            self.event.__str__(),
            self.status_type,
        )

    def __repr__(self):
        values = self.event.get_repr_values() + [self.status_type]
        return 'TimelineEventStatus({})'.format(
            ', '.join(repr(v) for v in values)
        )

    def get_project(self) -> Project:
        """Return the project for the event"""
        return self.event.project



"""Tests for email sending in the projectroles Django app"""

from django.conf import settings
from django.core import mail
from django.test import override_settings
from django.urls import reverse

from test_plus.test import TestCase, RequestFactory

from projectroles.app_settings import AppSettingAPI
from projectroles.models import (
    Project,
    SODARUser,
    SODAR_CONSTANTS,
    ROLE_RANKING,
)
from projectroles.email import (
    send_role_change_mail,
    send_project_leave_mail,
    send_invite_mail,
    send_invite_accept_mail,
    send_invite_expiry_mail,
    send_project_create_mail,
    send_project_move_mail,
    send_project_archive_mail,
    send_project_delete_mail,
    send_generic_mail,
    get_email_user,
    get_user_addr,
    get_email_footer,
    SETTINGS_LINK,
    SUBJECT_PREFIX,
)
from projectroles.tests.test_models import (
    ProjectMixin,
    RoleMixin,
    RoleAssignmentMixin,
    ProjectInviteMixin,
    SODARUserAdditionalEmailMixin,
)


app_settings = AppSettingAPI()


# SODAR constants
PROJECT_ROLE_OWNER = SODAR_CONSTANTS['PROJECT_ROLE_OWNER']
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']
PROJECT_ROLE_CONTRIBUTOR = SODAR_CONSTANTS['PROJECT_ROLE_CONTRIBUTOR']
PROJECT_ROLE_GUEST = SODAR_CONSTANTS['PROJECT_ROLE_GUEST']
PROJECT_TYPE_CATEGORY = SODAR_CONSTANTS['PROJECT_TYPE_CATEGORY']
PROJECT_TYPE_PROJECT = SODAR_CONSTANTS['PROJECT_TYPE_PROJECT']

# Local constants
APP_NAME = 'projectroles'
SUBJECT_BODY = 'Test subject'
MESSAGE_BODY = 'Test message'
INVALID_EMAIL = 'ahch0La8lo0eeT8u'
CUSTOM_HEADER = 'Custom header'
CUSTOM_FOOTER = 'Custom footer'

USER_ADD_EMAIL = 'user1@example.com'
USER_ADD_EMAIL2 = 'user2@example.com'


class TestEmailSending(
    ProjectMixin,
    RoleMixin,
    RoleAssignmentMixin,
    SODARUserAdditionalEmailMixin,
    ProjectInviteMixin,
    TestCase,
):
    """Tests for email sending"""

    def _get_settings_link(self) -> str:
        """Return message footer settings link"""
        return SETTINGS_LINK.format(
            url=self.request.build_absolute_uri(
                reverse('userprofile:settings_update')
            )
        )

    @classmethod
    def _get_project_modify_recipients(
        cls, project: Project, request_user: SODARUser
    ) -> list[SODARUser]:
        """
        Return owner and delegate users for project omitting request_user.
        Exclude inactive users but does NOT exclude app settings.
        """
        return [
            a.user
            for a in project.get_roles(
                max_rank=ROLE_RANKING[PROJECT_ROLE_DELEGATE]
            )
            if a.user.is_active and a.user != request_user
        ]

    @classmethod
    def _get_project_delete_recipients(
        cls, project: Project, request_user: SODARUser
    ) -> list[SODARUser]:
        """
        Return project users for project omitting request_user. Exclude inactive
        users but does NOT exclude app settings.
        """
        return [
            a.user
            for a in project.get_roles()
            if a.user.is_active and a.user != request_user
        ]

    def setUp(self):
        # Init roles
        self.init_roles()
        # Init users
        self.user_owner = self.make_user('owner')
        self.user_owner.email = 'owner_user@example.com'
        self.user_owner.save()
        self.user = self.make_user('user')
        self.user.email = 'user@example.com'
        self.user.save()
        self.user_inactive = self.make_user('user_inactive')
        self.user_inactive.email = 'user_inactive@example.com'
        self.user_inactive.is_active = False
        self.user_inactive.save()

        # Init projects
        self.category = self.make_project(
            'top_category', PROJECT_TYPE_CATEGORY, None
        )
        self.cat_owner_as = self.make_assignment(
            self.category, self.user_owner, self.role_owner
        )

        self.project = self.make_project(
            'sub_project', PROJECT_TYPE_PROJECT, self.category
        )
        self.owner_as = self.make_assignment(
            self.project, self.user_owner, self.role_owner
        )
        self.user_as = self.make_assignment(
            self.project, self.user, self.role_contributor
        )
        self.inactive_as = self.make_assignment(
            self.project, self.user_inactive, self.role_contributor
        )
        # Set up request factory and default request
        self.factory = RequestFactory()
        self.request = self.factory.get(reverse('home'))
        self.request.user = self.user_owner

    def test_send_role_change_mail_create(self):
        """Test send_role_change_mail() with role create"""
        email_sent = send_role_change_mail(
            change_type='create',
            project=self.project,
            user=self.user,
            role=self.user_as.role,
            request=self.request,
        )
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 1)
        self.assertEqual(mail.outbox[0].to[0], self.user.email)
        self.assertEqual(len(mail.outbox[0].reply_to), 1)
        self.assertEqual(mail.outbox[0].reply_to[0], self.user_owner.email)
        self.assertIn(SUBJECT_PREFIX, mail.outbox[0].subject)

    def test_send_role_change_mail_additional(self):
        """Test send_role_change_mail() with additional sender emails"""
        self.make_email(self.user, USER_ADD_EMAIL)
        self.make_email(self.user, USER_ADD_EMAIL2)
        email_sent = send_role_change_mail(
            change_type='create',
            project=self.project,
            user=self.user,
            role=self.user_as.role,
            request=self.request,
        )
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 3)
        self.assertEqual(
            mail.outbox[0].to,
            [self.user.email, USER_ADD_EMAIL, USER_ADD_EMAIL2],
        )
        self.assertEqual(len(mail.outbox[0].reply_to), 1)
        self.assertEqual(mail.outbox[0].reply_to[0], self.user_owner.email)

    def test_send_role_change_mail_additional_no_default(self):
        """Test send_role_change_mail() with additional sender emails and no default email"""
        self.make_email(self.user, USER_ADD_EMAIL)
        self.make_email(self.user, USER_ADD_EMAIL2)
        self.user.email = ''
        self.user.save()
        email_sent = send_role_change_mail(
            change_type='create',
            project=self.project,
            user=self.user,
            role=self.user_as.role,
            request=self.request,
        )
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 2)
        self.assertEqual(
            mail.outbox[0].to,
            [USER_ADD_EMAIL, USER_ADD_EMAIL2],
        )
        self.assertEqual(len(mail.outbox[0].reply_to), 1)
        self.assertEqual(mail.outbox[0].reply_to[0], self.user_owner.email)

    def test_send_role_change_mail_additional_reply(self):
        """Test send_role_change_mail() with additional reply-to emails"""
        self.make_email(self.user_owner, USER_ADD_EMAIL)
        self.make_email(self.user_owner, USER_ADD_EMAIL2)
        email_sent = send_role_change_mail(
            change_type='create',
            project=self.project,
            user=self.user,
            role=self.user_as.role,
            request=self.request,
        )
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 1)
        self.assertEqual(
            mail.outbox[0].to,
            [self.user.email],
        )
        self.assertEqual(len(mail.outbox[0].reply_to), 3)
        self.assertEqual(
            mail.outbox[0].reply_to,
            [self.user_owner.email, USER_ADD_EMAIL, USER_ADD_EMAIL2],
        )

    def test_send_role_change_mail_update(self):
        """Test send_role_change_mail() with role update"""
        email_sent = send_role_change_mail(
            change_type='update',
            project=self.project,
            user=self.user,
            role=self.user_as.role,
            request=self.request,
        )
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 1)
        self.assertEqual(mail.outbox[0].to[0], self.user.email)
        self.assertEqual(len(mail.outbox[0].reply_to), 1)
        self.assertEqual(mail.outbox[0].reply_to[0], self.user_owner.email)

    def test_send_role_change_mail_delete(self):
        """Test send_role_change_mail() with role delete"""
        email_sent = send_role_change_mail(
            change_type='delete',
            project=self.project,
            user=self.user,
            role=None,
            request=self.request,
        )
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 1)
        self.assertEqual(mail.outbox[0].to[0], self.user.email)
        self.assertEqual(len(mail.outbox[0].reply_to), 1)
        self.assertEqual(mail.outbox[0].reply_to[0], self.user_owner.email)

    def test_send_invite_mail(self):
        """Test send_invite_mail()"""
        invite = self.make_invite(
            email=self.user.email,
            project=self.project,
            role=self.role_guest,
            issuer=self.user_owner,
        )
        email_sent = send_invite_mail(invite, self.request)
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 1)
        self.assertEqual(mail.outbox[0].to[0], self.user.email)
        self.assertEqual(len(mail.outbox[0].reply_to), 1)
        self.assertEqual(mail.outbox[0].reply_to[0], self.user_owner.email)
        self.assertIn(SUBJECT_PREFIX, mail.outbox[0].subject)

    def test_send_invite_accept_mail(self):
        """Test send_invite_accept_mail()"""
        invite = self.make_invite(
            email=self.user.email,
            project=self.project,
            role=self.role_guest,
            issuer=self.user_owner,
        )
        email_sent = send_invite_accept_mail(invite, self.request, self.user)
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 1)
        self.assertEqual(mail.outbox[0].to[0], self.user_owner.email)
        self.assertEqual(len(mail.outbox[0].reply_to), 0)
        self.assertIn(SUBJECT_PREFIX, mail.outbox[0].subject)

    def test_send_invite_expiry_mail(self):
        """Test send_invite_expiry_mail()"""
        invite = self.make_invite(
            email=self.user.email,
            project=self.project,
            role=self.role_guest,
            issuer=self.user_owner,
        )
        email_sent = send_invite_expiry_mail(
            invite, self.request, self.user.username
        )
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 1)
        self.assertEqual(mail.outbox[0].to[0], self.user_owner.email)
        self.assertEqual(len(mail.outbox[0].reply_to), 0)
        self.assertIn(SUBJECT_PREFIX, mail.outbox[0].subject)

    def test_send_project_leave_mail(self):
        """Test send_project_move_mail()"""
        email_sent = send_project_leave_mail(
            project=self.project,
            user=self.user,
            request=self.request,
        )
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 1)
        self.assertEqual(mail.outbox[0].to[0], self.user_owner.email)
        self.assertEqual(len(mail.outbox[0].reply_to), 0)
        self.assertIn(SUBJECT_PREFIX, mail.outbox[0].subject)

    def test_send_project_create_mail(self):
        """Test send_project_create_mail()"""
        new_project = self.make_project(
            'New Project', PROJECT_TYPE_PROJECT, self.category
        )
        self.make_assignment(new_project, self.user, self.role_owner)
        self.request.user = self.user
        email_sent = send_project_create_mail(
            project=new_project,
            recipients=self._get_project_modify_recipients(
                new_project, self.user
            ),
            request=self.request,
        )
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 1)
        self.assertEqual(mail.outbox[0].to[0], self.user_owner.email)
        self.assertEqual(len(mail.outbox[0].reply_to), 1)
        self.assertEqual(mail.outbox[0].reply_to[0], self.user.email)
        self.assertIn(SUBJECT_PREFIX, mail.outbox[0].subject)

    def test_send_project_create_mail_parent_delegate(self):
        """Test send_project_create_mail() with parent delegate"""
        cat_delegate = self.make_user('cat_delegate')
        cat_delegate.email = 'cat_delegate@example.com'
        cat_delegate.save()
        self.make_assignment(self.category, cat_delegate, self.role_delegate)
        new_project = self.make_project(
            'New Project', PROJECT_TYPE_PROJECT, self.category
        )
        self.make_assignment(new_project, self.user, self.role_owner)
        self.request.user = self.user
        email_sent = send_project_create_mail(
            project=new_project,
            recipients=self._get_project_modify_recipients(
                new_project, self.user
            ),
            request=self.request,
        )
        self.assertEqual(email_sent, 2)
        self.assertEqual(len(mail.outbox), 2)
        self.assertEqual(mail.outbox[0].to[0], cat_delegate.email)
        self.assertEqual(mail.outbox[1].to[0], self.user_owner.email)

    def test_send_project_create_mail_inactive_parent_owner(self):
        """Test send_project_create_mail() with inactive parent owner"""
        self.user_owner.is_active = False
        self.user_owner.save()
        new_project = self.make_project(
            'New Project', PROJECT_TYPE_PROJECT, self.category
        )
        self.make_assignment(new_project, self.user, self.role_owner)
        self.request.user = self.user
        email_sent = send_project_create_mail(
            project=new_project,
            recipients=self._get_project_modify_recipients(
                new_project, self.user
            ),
            request=self.request,
        )
        self.assertEqual(email_sent, 0)

    def test_send_project_create_mail_disable_notify(self):
        """Test send_project_create_mail() with disabled email notifications"""
        app_settings.set(
            APP_NAME, 'notify_email_project', False, user=self.user_owner
        )
        new_project = self.make_project(
            'New Project', PROJECT_TYPE_PROJECT, self.category
        )
        self.make_assignment(new_project, self.user, self.role_owner)
        self.request.user = self.user
        email_sent = send_project_create_mail(
            project=new_project,
            recipients=self._get_project_modify_recipients(
                new_project, self.user
            ),
            request=self.request,
        )
        self.assertEqual(email_sent, 0)

    def test_send_project_move_mail(self):
        """Test send_project_move_mail()"""
        self.request.user = self.user
        email_sent = send_project_move_mail(
            project=self.project,
            recipients=self._get_project_modify_recipients(
                self.project, self.user
            ),
            request=self.request,
        )
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 1)
        self.assertEqual(mail.outbox[0].to[0], self.user_owner.email)
        self.assertEqual(len(mail.outbox[0].reply_to), 1)
        self.assertEqual(mail.outbox[0].reply_to[0], self.user.email)
        self.assertIn(SUBJECT_PREFIX, mail.outbox[0].subject)

    def test_send_project_move_mail_parent_delegate(self):
        """Test send_project_move_mail() with parent delegate"""
        cat_delegate = self.make_user('cat_delegate')
        cat_delegate.email = 'cat_delegate@example.com'
        cat_delegate.save()
        self.make_assignment(self.category, cat_delegate, self.role_delegate)
        new_project = self.make_project(
            'New Project', PROJECT_TYPE_PROJECT, self.category
        )
        self.make_assignment(new_project, self.user, self.role_owner)
        self.request.user = self.user
        email_sent = send_project_move_mail(
            project=new_project,
            recipients=self._get_project_modify_recipients(
                new_project, self.user
            ),
            request=self.request,
        )
        self.assertEqual(email_sent, 2)
        self.assertEqual(len(mail.outbox), 2)
        self.assertEqual(mail.outbox[0].to[0], cat_delegate.email)
        self.assertEqual(mail.outbox[1].to[0], self.user_owner.email)

    def test_send_project_move_mail_disable_notify(self):
        """Test send_project_move_mail() with disabled email notifications"""
        app_settings.set(
            APP_NAME, 'notify_email_project', False, user=self.user_owner
        )
        new_project = self.make_project(
            'New Project', PROJECT_TYPE_PROJECT, self.category
        )
        self.make_assignment(new_project, self.user, self.role_owner)
        self.request.user = self.user
        email_sent = send_project_move_mail(
            project=new_project,
            recipients=self._get_project_modify_recipients(
                new_project, self.user
            ),
            request=self.request,
        )
        self.assertEqual(email_sent, 0)

    def test_send_project_move_mail_inactive_parent_owner(self):
        """Test send_project_move_mail() with inactive parent owner"""
        self.user_owner.is_active = False
        self.user_owner.save()
        self.request.user = self.user
        email_sent = send_project_move_mail(
            project=self.project,
            recipients=self._get_project_modify_recipients(
                self.project, self.user
            ),
            request=self.request,
        )
        self.assertEqual(email_sent, 0)

    def test_send_project_archive_mail(self):
        """Test send_project_archive_mail()"""
        email_sent = send_project_archive_mail(
            project=self.project,
            action='archive',
            request=self.request,
        )
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 1)
        self.assertEqual(mail.outbox[0].to[0], self.user.email)
        self.assertEqual(len(mail.outbox[0].reply_to), 0)
        self.assertIn(SUBJECT_PREFIX, mail.outbox[0].subject)

    def test_send_project_archive_mail_disable_notify(self):
        """Test send_project_archive_mail() with disabled email notifications"""
        app_settings.set(
            APP_NAME, 'notify_email_project', False, user=self.user
        )
        email_sent = send_project_archive_mail(
            project=self.project,
            action='archive',
            request=self.request,
        )
        self.assertEqual(email_sent, 0)

    def test_send_project_archive_mail_inactive_recipient(self):
        """Test send_project_archive_mail() with inactive recipient"""
        self.user.is_active = False
        self.user.save()
        email_sent = send_project_archive_mail(
            project=self.project,
            action='archive',
            request=self.request,
        )
        self.assertEqual(email_sent, 0)

    def test_send_project_delete_mail(self):
        """Test send_project_delete_mail()"""
        email_sent = send_project_delete_mail(
            project=self.project,
            recipients=self._get_project_delete_recipients(
                self.project, self.user_owner
            ),
            request=self.request,
        )
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 1)
        self.assertEqual(mail.outbox[0].to[0], self.user.email)
        self.assertEqual(len(mail.outbox[0].reply_to), 0)
        self.assertIn(SUBJECT_PREFIX, mail.outbox[0].subject)

    def test_send_project_delete_mail_disable_notify(self):
        """Test send_project_delete_mail() with disabled email notifications"""
        app_settings.set(
            APP_NAME, 'notify_email_project', False, user=self.user
        )
        email_sent = send_project_delete_mail(
            project=self.project,
            recipients=self._get_project_delete_recipients(
                self.project, self.user_owner
            ),
            request=self.request,
        )
        self.assertEqual(email_sent, 0)

    def test_send_project_delete_mail_inactive_recipient(self):
        """Test send_project_delete_mail() with inactive recipient"""
        self.user.is_active = False
        self.user.save()
        email_sent = send_project_delete_mail(
            project=self.project,
            recipients=self._get_project_delete_recipients(
                self.project, self.user_owner
            ),
            request=self.request,
        )
        self.assertEqual(email_sent, 0)

    def test_send_generic_mail_user(self):
        """Test send_generic_mail() with User recipient"""
        email_sent = send_generic_mail(
            subject_body=SUBJECT_BODY,
            message_body=MESSAGE_BODY,
            recipient_list=[self.user],
            request=self.request,
            reply_to=[self.user_owner.email],
        )
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 1)
        self.assertEqual(mail.outbox[0].to[0], self.user.email)
        self.assertEqual(len(mail.outbox[0].reply_to), 1)
        self.assertEqual(mail.outbox[0].reply_to[0], self.user_owner.email)
        self.assertIn(SUBJECT_PREFIX, mail.outbox[0].subject)

    def test_send_generic_mail_user_additional(self):
        """Test send_generic_mail() with User and additional emails"""
        self.make_email(self.user, USER_ADD_EMAIL)
        self.make_email(self.user, USER_ADD_EMAIL2)
        email_sent = send_generic_mail(
            subject_body=SUBJECT_BODY,
            message_body=MESSAGE_BODY,
            recipient_list=[self.user],
            request=self.request,
            reply_to=[self.user_owner.email],
        )
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 3)
        self.assertEqual(
            mail.outbox[0].to,
            [self.user.email, USER_ADD_EMAIL, USER_ADD_EMAIL2],
        )
        self.assertEqual(len(mail.outbox[0].reply_to), 1)
        self.assertEqual(mail.outbox[0].reply_to[0], self.user_owner.email)

    def test_send_generic_mail_str(self):
        """Test send_generic_mail() with email string recipient"""
        email_sent = send_generic_mail(
            subject_body=SUBJECT_BODY,
            message_body=MESSAGE_BODY,
            recipient_list=[self.user.email],
            request=self.request,
            reply_to=[self.user_owner.email],
        )
        self.assertEqual(email_sent, 1)
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(len(mail.outbox[0].to), 1)
        self.assertEqual(mail.outbox[0].to[0], self.user.email)
        self.assertEqual(len(mail.outbox[0].reply_to), 1)
        self.assertEqual(mail.outbox[0].reply_to[0], self.user_owner.email)

    def test_send_generic_mail_multiple(self):
        """Test send_generic_mail() with multiple recipients"""
        user_new = self.make_user('newuser')
        email_sent = send_generic_mail(
            subject_body=SUBJECT_BODY,
            message_body=MESSAGE_BODY,
            recipient_list=[self.user_owner, user_new],
            request=self.request,
        )
        self.assertEqual(email_sent, 2)
        self.assertEqual(len(mail.outbox), 2)

    def test_get_email_footer(self):
        """Test get_email_footer() with default admin"""
        footer = get_email_footer(self.request)
        self.assertIn(
            settings.PROJECTROLES_SUPPORT_CONTACT.split(':')[0], footer
        )
        self.assertIn(
            settings.PROJECTROLES_SUPPORT_CONTACT.split(':')[1], footer
        )
        self.assertIn(self._get_settings_link(), footer)

    @override_settings(PROJECTROLES_SUPPORT_CONTACT=None)
    def test_get_email_footer_no_support_contact(self):
        """Test get_email_footer() with no support contact"""
        footer = get_email_footer(self.request)
        self.assertIn(settings.ADMINS[0][0], footer)
        self.assertIn(settings.ADMINS[0][1], footer)
        self.assertIn(self._get_settings_link(), footer)

    @override_settings(ADMINS=[])
    def test_get_email_footer_no_admin(self):
        """Test get_email_footer() with no admins"""
        footer = get_email_footer(self.request)
        self.assertIn(
            settings.PROJECTROLES_SUPPORT_CONTACT.split(':')[0], footer
        )
        self.assertIn(
            settings.PROJECTROLES_SUPPORT_CONTACT.split(':')[1], footer
        )
        self.assertIn(self._get_settings_link(), footer)

    @override_settings(ADMINS=[], PROJECTROLES_SUPPORT_CONTACT=None)
    def test_get_email_footer_no_admin_no_support(self):
        """Test get_email_footer() with no admins and no support contact"""
        self.assertEqual(
            get_email_footer(self.request), self._get_settings_link()
        )

    def test_get_email_footer_no_settings(self):
        """Test get_email_footer() with no settings link"""
        footer = get_email_footer(self.request, settings_link=False)
        self.assertIn(
            settings.PROJECTROLES_SUPPORT_CONTACT.split(':')[0], footer
        )
        self.assertIn(
            settings.PROJECTROLES_SUPPORT_CONTACT.split(':')[1], footer
        )
        self.assertNotIn(self._get_settings_link(), footer)

    @override_settings(PROJECTROLES_EMAIL_HEADER=CUSTOM_HEADER)
    def test_custom_header(self):
        """Test send_generic_mail() with custom header"""
        send_generic_mail(
            subject_body=SUBJECT_BODY,
            message_body=MESSAGE_BODY,
            recipient_list=[self.user_owner.email],
            request=self.request,
            reply_to=[self.user_owner.email],
        )
        self.assertIn(CUSTOM_HEADER, mail.outbox[0].body)
        self.assertNotIn(CUSTOM_FOOTER, mail.outbox[0].body)

    @override_settings(PROJECTROLES_EMAIL_FOOTER=CUSTOM_FOOTER)
    def test_custom_footer(self):
        """Test send_generic_mail() with custom footer"""
        send_generic_mail(
            subject_body=SUBJECT_BODY,
            message_body=MESSAGE_BODY,
            recipient_list=[self.user_owner.email],
            request=self.request,
            reply_to=[self.user_owner.email],
        )
        self.assertNotIn(CUSTOM_HEADER, mail.outbox[0].body)
        self.assertIn(CUSTOM_FOOTER, mail.outbox[0].body)

    def test_get_email_user(self):
        """Test get_email_user()"""
        self.assertEqual(
            get_email_user(self.user_owner), 'owner (owner_user@example.com)'
        )

    def test_get_email_user_no_email(self):
        """Test get_email_user() without email"""
        self.user_owner.email = ''
        self.assertEqual(get_email_user(self.user_owner), 'owner')

    def test_get_email_user_name(self):
        """Test get_email_user() with name"""
        self.user_owner.name = 'Owner User'
        self.assertEqual(
            get_email_user(self.user_owner),
            'Owner User (owner_user@example.com)',
        )

    def test_get_email_user_first_last_name(self):
        """Test get_email_user() with first and last name"""
        self.user_owner.first_name = 'Owner'
        self.user_owner.last_name = 'User'
        self.assertEqual(
            get_email_user(self.user_owner),
            'Owner User (owner_user@example.com)',
        )

    def test_get_user_addr(self):
        """Test get_user_addr() with standard user email"""
        self.assertEqual(get_user_addr(self.user), [self.user.email])

    def test_get_user_addr_additional(self):
        """Test get_user_addr() with additional user emails"""
        self.make_email(self.user, USER_ADD_EMAIL)
        self.make_email(self.user, USER_ADD_EMAIL2)
        self.assertEqual(
            get_user_addr(self.user),
            [self.user.email, USER_ADD_EMAIL, USER_ADD_EMAIL2],
        )

    def test_get_user_addr_additional_unverified(self):
        """Test get_user_addr() with additional and unverified user emails"""
        self.make_email(self.user, USER_ADD_EMAIL)
        self.make_email(self.user, USER_ADD_EMAIL2, verified=False)
        self.assertEqual(
            get_user_addr(self.user), [self.user.email, USER_ADD_EMAIL]
        )

    def test_get_user_addr_additional_no_default(self):
        """Test get_user_addr() with additional user emails and no default"""
        self.make_email(self.user, USER_ADD_EMAIL)
        self.make_email(self.user, USER_ADD_EMAIL2)
        self.user.email = ''
        self.assertEqual(
            get_user_addr(self.user), [USER_ADD_EMAIL, USER_ADD_EMAIL2]
        )



"""Context processors for the projectroles app"""

from math import ceil

from django.conf import settings

from projectroles.plugins import PluginAPI
from projectroles.urls import urlpatterns
from projectroles.app_links import ROLE_URLS


plugin_api = PluginAPI()


# Local constants
SIDEBAR_ICON_MIN_SIZE = 18
SIDEBAR_ICON_MAX_SIZE = 42


def urls_processor(request):
    """Context processor for providing projectroles URLs for the sidebar"""
    return {
        'projectroles_urls': urlpatterns,
        'role_urls': ROLE_URLS,
    }


def site_app_processor(request):
    """
    Context processor for providing site apps for the site titlebar dropdown.
    """
    site_apps = plugin_api.get_active_plugins('site_app')
    return {
        'site_apps': [
            a
            for a in site_apps
            if not a.app_permission or request.user.has_perm(a.app_permission)
        ],
    }


def app_alerts_processor(request):
    """
    Context processor for checking app alert status.
    """
    if request.user and request.user.is_authenticated:
        app_alerts = plugin_api.get_backend_api('appalerts_backend')
        if app_alerts:
            return {
                'app_alerts': app_alerts.get_model()
                .objects.filter(user=request.user, active=True)
                .count()
            }
    return {'app_alerts': 0}


def sidebar_processor(request):
    """
    Context processor for providing sidebar information.
    """
    sidebar_icon_size = sorted(
        [
            SIDEBAR_ICON_MIN_SIZE,
            getattr(settings, 'PROJECTROLES_SIDEBAR_ICON_SIZE', 32),
            SIDEBAR_ICON_MAX_SIZE,
        ]
    )[1]
    return {
        'sidebar_icon_size': sidebar_icon_size,
        'sidebar_notch_pos': ceil(sidebar_icon_size / 3),
        'sidebar_notch_size': min(ceil(sidebar_icon_size / 2), 12),
        'sidebar_padding': ceil(sidebar_icon_size / 4.5),
    }



from django.urls import path

from sodarcache import views_api

app_name = 'sodarcache'

urlpatterns = [
    path(
        route='api/retrieve/<uuid:project>/<str:app_name>/<str:item_name>',
        view=views_api.CacheItemRetrieveAPIView.as_view(),
        name='api_retrieve',
    ),
    path(
        route='api/retrieve/date/<uuid:project>/<str:app_name>/<str:item_name>',
        view=views_api.CacheItemDateRetrieveAPIView.as_view(),
        name='api_retrieve_date',
    ),
    path(
        route='api/set/<uuid:project>/<str:app_name>/<str:item_name>',
        view=views_api.CacheItemSetAPIView.as_view(),
        name='api_set',
    ),
]



"""Tests for UI view permissions in the adminalerts app"""

from django.test import override_settings
from django.urls import reverse

# Projectroles dependency
from projectroles.tests.test_permissions import SiteAppPermissionTestBase

from adminalerts.tests.test_models import AdminAlertMixin


class AdminalertsPermissionTestBase(AdminAlertMixin, SiteAppPermissionTestBase):
    """Base test class for adminalerts UI view permission tests"""

    def setUp(self):
        super().setUp()
        # Create alert
        self.alert = self.make_alert(
            message='alert',
            user=self.superuser,
            description='description',
            active=True,
        )


class TestAdminAlertListView(AdminalertsPermissionTestBase):
    """Permission tests for AdminAlertListView"""

    def setUp(self):
        super().setUp()
        self.url = reverse('adminalerts:list')

    def test_get(self):
        """Test AdminAlertListView GET"""
        good_users = [self.superuser]
        bad_users = [self.anonymous, self.regular_user]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        good_users = [self.superuser]
        bad_users = [self.anonymous, self.regular_user]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)


class TestAdminAlertDetailView(AdminalertsPermissionTestBase):
    """Permission tests for dminAlertDetailView"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'adminalerts:detail', kwargs={'adminalert': self.alert.sodar_uuid}
        )

    def test_get(self):
        """Test AdminAlertDetailView GET"""
        good_users = [self.superuser, self.regular_user]
        bad_users = [self.anonymous]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        good_users = [self.superuser, self.regular_user]
        bad_users = [self.anonymous]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)


class TestAdminAlertCreateView(AdminalertsPermissionTestBase):
    """Permission tests for AdminAlertCreateView"""

    def setUp(self):
        super().setUp()
        self.url = reverse('adminalerts:create')

    def test_get(self):
        """Test AdminAlertCreateView GET"""
        good_users = [self.superuser]
        bad_users = [self.anonymous, self.regular_user]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        good_users = [self.superuser]
        bad_users = [self.anonymous, self.regular_user]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)


class TestAdminAlertUpdateView(AdminalertsPermissionTestBase):
    """Permission tests for AdminAlertUpdateView"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'adminalerts:update', kwargs={'adminalert': self.alert.sodar_uuid}
        )

    def test_get(self):
        """Test AdminAlertUpdateView GET"""
        good_users = [self.superuser]
        bad_users = [self.anonymous, self.regular_user]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        good_users = [self.superuser]
        bad_users = [self.anonymous, self.regular_user]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)


class TestAdminAlertDeleteView(AdminalertsPermissionTestBase):
    """Permission tests for AdminAlertDeleteView"""

    def setUp(self):
        super().setUp()
        self.url = reverse(
            'adminalerts:delete', kwargs={'adminalert': self.alert.sodar_uuid}
        )

    def test_get(self):
        """Test AdminAlertDeleteView GET"""
        good_users = [self.superuser]
        bad_users = [self.anonymous, self.regular_user]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)

    @override_settings(PROJECTROLES_ALLOW_ANONYMOUS=True)
    def test_get_anon(self):
        """Test GET with anonymous access"""
        good_users = [self.superuser]
        bad_users = [self.anonymous, self.regular_user]
        self.assert_response(self.url, good_users, 200)
        self.assert_response(self.url, bad_users, 302)



from django import forms

# Projectroles dependency
from projectroles.forms import SODARForm


class UserTokenCreateForm(SODARForm):
    """This form allows token creation"""

    #: Time to live in hours
    ttl = forms.IntegerField(
        label='Time to live',
        min_value=0,
        required=True,
        initial=0,
        help_text='Time to live in hours, set to 0 for tokens that never '
        'expire.',
    )



"""Email creation and sending for the projectroles app"""

import logging
import re

from typing import Optional

from django.conf import settings
from django.contrib import auth, messages
from django.core.mail import EmailMessage
from django.http import HttpRequest
from django.urls import reverse
from django.utils.timezone import localtime

from projectroles.app_settings import AppSettingAPI
from projectroles.models import (
    Project,
    Role,
    ProjectInvite,
    SODARUser,
    SODARUserAdditionalEmail,
    SODAR_CONSTANTS,
    ROLE_RANKING,
)
from projectroles.plugins import PluginAPI
from projectroles.utils import get_display_name


app_settings = AppSettingAPI()
logger = logging.getLogger(__name__)
plugin_api = PluginAPI()
User = auth.get_user_model()


# Settings
SUBJECT_PREFIX = settings.EMAIL_SUBJECT_PREFIX.strip() + ' '
EMAIL_SENDER = settings.EMAIL_SENDER
DEBUG = settings.DEBUG
SITE_TITLE = settings.SITE_INSTANCE_TITLE

# SODAR constants
PROJECT_ROLE_DELEGATE = SODAR_CONSTANTS['PROJECT_ROLE_DELEGATE']

# Local constants
APP_NAME = 'projectroles'
EMAIL_RE = re.compile(r'(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)')


# Generic Elements -------------------------------------------------------------


MESSAGE_HEADER = r'''
Dear {recipient},

This email has been automatically sent to you by {site_title}.

'''.lstrip()

MESSAGE_HEADER_NO_RECIPIENT = r'''
This email has been automatically sent to you by {site_title}.
'''.lstrip()

NO_REPLY_NOTE = r'''
Please do not reply to this email.
'''

MESSAGE_FOOTER = r'''

For support and reporting issues regarding {site_title},
contact {admin_name} ({admin_email}).
'''

SETTINGS_LINK = r'''
You can manage receiving of automated emails in your user settings:
{url}
'''


# Role Change Template ---------------------------------------------------------


SUBJECT_ROLE_CREATE = 'Membership granted for {project_label} "{project}"'
SUBJECT_ROLE_UPDATE = 'Membership changed in {project_label} "{project}"'
SUBJECT_ROLE_DELETE = 'Membership removed from {project_label} "{project}"'
SUBJECT_ROLE_LEAVE = 'Member {user_name} left {project_label} "{project}"'

MESSAGE_ROLE_CREATE = r'''
{issuer} has granted you the membership
in {project_label} "{project}" with the role of "{role}".

To access the {project_label} in {site_title}, please click on
the following link:
{project_url}
'''.lstrip()

MESSAGE_ROLE_UPDATE = r'''
{issuer} has changed your membership
role in {project_label} "{project}" into "{role}".

To access the {project_label} in {site_title}, please click on
the following link:
{project_url}
'''.lstrip()

MESSAGE_ROLE_DELETE = r'''
{issuer} has removed your membership from {project_label} "{project}".
'''.lstrip()

MESSAGE_ROLE_LEAVE = r'''
Member {user_name} has left the {project_label} "{project}".
For them to regain access, it has to be granted again by {project_label}
owner or delegate.
'''.lstrip()


# Invite Template --------------------------------------------------------------


SUBJECT_INVITE = 'Invitation for {project_label} "{project}"'

MESSAGE_INVITE_BODY = r'''
You have been invited by {issuer}
to share data in the {project_label} "{project}" with the
role of "{role}".

To accept the invitation and access the {project_label} in {site_title},
please click on the following link:
{invite_url}

Please note that the link is only to be used once. After successfully
accepting the invitation, please access the project with its URL or
through the project list on the site's "Home" page.

This invitation will expire on {date_expire}.
'''

MESSAGE_INVITE_ISSUER = r'''
Message from the sender of this invitation:
----------------------------------------
{}
----------------------------------------
'''


# Invite Acceptance Notification Template --------------------------------------


SUBJECT_ACCEPT = 'Invitation accepted by {user} for {project_label} "{project}"'

MESSAGE_ACCEPT_BODY = r'''
Invitation sent by you for role of "{role}" in {project_label} "{project}"
has been accepted by {user}.
They have been granted access in the {project_label} accordingly.
'''.lstrip()


# Invite Expiry Notification Template ------------------------------------------


SUBJECT_EXPIRY = 'Expired invitation used by {user_name} in "{project}"'

MESSAGE_EXPIRY_BODY = r'''
Invitation sent by you for role of "{role}" in {project_label} "{project}"
was attempted to be used by {user_name} ({user_email}).

This invitation has expired on {date_expire}. Because of this,
access was not granted to the user.

Please add the role manually with "Add Member", if you still wish
to grant the user access to the {project_label}.
'''.lstrip()


# Project/Category Creation Notification Template ------------------------------


SUBJECT_PROJECT_CREATE = '{project_type} "{project}" created by {user}'

MESSAGE_PROJECT_CREATE_BODY = r'''
{user} has created a new {project_type}
under "{category}".
You are receiving this email because you are the owner of the parent category.
You have automatically inherited owner rights to the created {project_type}.

Title: {project}
Owner: {owner}

You can access the project at the following link:
{project_url}
'''.lstrip()


# Project/Category Moving Notification Template ------------------------------


SUBJECT_PROJECT_MOVE = '{project_type} "{project}" moved by {user}'

MESSAGE_PROJECT_MOVE_BODY = r'''
{user} has moved the {project_type} "{project}"
under "{category}".
You are receiving this email because you are the owner of the parent category.
You have automatically inherited owner rights to the created {project_type}.

Title: {project}
Owner: {owner}

You can access the project at the following link:
{project_url}
'''.lstrip()


# Project Archiving Template ---------------------------------------------------


SUBJECT_PROJECT_ARCHIVE = '{project_label_title} "{project}" archived by {user}'

MESSAGE_PROJECT_ARCHIVE = r'''
{user} has archived "{project}".
The {project_label} is now read-only. Modifying data in
{project_label} apps has been disabled. Existing project data
can still be accessed and user roles can be modified.

You can access the archived {project_label} at the following link:
{project_url}
'''.lstrip()


# Project Unarchiving Template -------------------------------------------------


SUBJECT_PROJECT_UNARCHIVE = (
    '{project_label_title} "{project}" unarchived by {user}'
)

MESSAGE_PROJECT_UNARCHIVE = r'''
{user} has unarchived "{project}".
{project_label_title} data can now be modified and write
access for users has been restored.

You can access the {project_label} at the following link:
{project_url}
'''.lstrip()


# Project Deletion Template ----------------------------------------------------

SUBJECT_PROJECT_DELETE = '{project_label_title} "{project}" deleted by {user}'

MESSAGE_PROJECT_DELETE = r'''
{user} has deleted "{project}".
The {project_label} has been removed from the site and can no
longer be accessed.
'''.lstrip()


# Email composing helpers ------------------------------------------------------


def get_email_user(user: SODARUser) -> str:
    """
    Return a string representation of a user object for emails.

    :param user: SODARUser object
    :return: string
    """
    ret = user.get_full_name()
    if user.email:
        ret += f' ({user.email})'
    return ret


def get_invite_body(
    project: Project,
    issuer: SODARUser,
    role_name: str,
    invite_url: str,
    date_expire_str: str,
) -> str:
    """
    Return the invite content header.

    :param project: Project object
    :param issuer: User object
    :param role_name: Display name of the Role object
    :param invite_url: Generated URL for the invite
    :param date_expire_str: Expiry date as a pre-formatted string
    :return: string
    """
    body = MESSAGE_HEADER_NO_RECIPIENT.format(site_title=SITE_TITLE)
    body += MESSAGE_INVITE_BODY.format(
        issuer=get_email_user(issuer),
        project=project.title,
        role=role_name,
        invite_url=invite_url,
        date_expire=date_expire_str,
        site_title=SITE_TITLE,
        project_label=get_display_name(project.type),
    )
    if not issuer.email and not settings.PROJECTROLES_EMAIL_SENDER_REPLY:
        body += NO_REPLY_NOTE
    return body


def get_invite_message(message: Optional[str] = None) -> str:
    """
    Return the message from invite issuer, of empty string if not set.

    :param message: Optional user message as string
    :return: string
    """
    if message and len(message) > 0:
        return MESSAGE_INVITE_ISSUER.format(message)
    return ''


def get_email_header(header: str) -> str:
    """
    Return the email header.

    :param header: Header string
    :return: string
    """
    return getattr(settings, 'PROJECTROLES_EMAIL_HEADER', None) or header


def get_email_footer(request: HttpRequest, settings_link: bool = True) -> str:
    """
    Return the email footer.

    :param request: HttpRequest object
    :param settings_link: Include link to user settings if True (optional)
    :return: string
    """
    footer = ''
    custom_footer = getattr(settings, 'PROJECTROLES_EMAIL_FOOTER', None)
    support_contact = getattr(settings, 'PROJECTROLES_SUPPORT_CONTACT', None)
    admin_recipient = settings.ADMINS[0] if settings.ADMINS else None
    if custom_footer:
        footer += '\n' + custom_footer
    elif support_contact and ':' in support_contact:
        footer += MESSAGE_FOOTER.format(
            site_title=SITE_TITLE,
            admin_name=support_contact.split(':')[0],
            admin_email=support_contact.split(':')[1],
        )
    elif admin_recipient:
        footer += MESSAGE_FOOTER.format(
            site_title=SITE_TITLE,
            admin_name=admin_recipient[0],
            admin_email=admin_recipient[1],
        )
    # Add user settings link if enabled and userprofile app is active
    if request and settings_link and plugin_api.get_app_plugin('userprofile'):
        footer += SETTINGS_LINK.format(
            url=request.build_absolute_uri(
                reverse('userprofile:settings_update')
            )
        )
    return footer


def get_invite_subject(project: Project) -> str:
    """
    Return invite email subject.

    :param project: Project object
    :return: string
    """
    return SUBJECT_PREFIX + SUBJECT_INVITE.format(
        project=project.title, project_label=get_display_name(project.type)
    )


def get_role_change_subject(change_type: str, project: Project) -> str:
    """
    Return role change email subject.

    :param change_type: Change type ('create', 'update', 'delete')
    :param project: Project object
    :return: String
    """
    subject = SUBJECT_PREFIX
    subject_kwargs = {
        'project': project.title,
        'project_label': get_display_name(project.type),
    }
    if change_type == 'create':
        subject += SUBJECT_ROLE_CREATE.format(**subject_kwargs)
    elif change_type == 'update':
        subject += SUBJECT_ROLE_UPDATE.format(**subject_kwargs)
    elif change_type == 'delete':
        subject += SUBJECT_ROLE_DELETE.format(**subject_kwargs)
    return subject


def get_role_change_body(
    change_type: str,
    project: Project,
    user_name: str,
    role_name: str,
    issuer: SODARUser,
    request: HttpRequest,
) -> str:
    """
    Return role change email body.

    :param change_type: Change type ('create', 'update', 'delete')
    :param project: Project object
    :param user_name: Name of target user
    :param role_name: Name of role as string
    :param issuer: User object for issuing user
    :param request: HttpRequest object
    :return: String
    """
    project_url = request.build_absolute_uri(
        reverse('projectroles:detail', kwargs={'project': project.sodar_uuid})
    )
    body = get_email_header(
        MESSAGE_HEADER.format(recipient=user_name, site_title=SITE_TITLE)
    )
    if change_type == 'create':
        body += MESSAGE_ROLE_CREATE.format(
            issuer=get_email_user(issuer),
            role=role_name,
            project=project.title,
            project_url=project_url,
            site_title=SITE_TITLE,
            project_label=get_display_name(project.type),
        )

    elif change_type == 'update':
        body += MESSAGE_ROLE_UPDATE.format(
            issuer=get_email_user(issuer),
            role=role_name,
            project=project.title,
            project_url=project_url,
            site_title=SITE_TITLE,
            project_label=get_display_name(project.type),
        )

    elif change_type == 'delete':
        body += MESSAGE_ROLE_DELETE.format(
            issuer=get_email_user(issuer),
            project=project.title,
            project_label=get_display_name(project.type),
        )
    if not issuer.email and not settings.PROJECTROLES_EMAIL_SENDER_REPLY:
        body += NO_REPLY_NOTE
    body += get_email_footer(request)
    return body


def get_user_addr(user: SODARUser) -> list[str]:
    """
    Return all the email addresses for a user as a list. Verified emails set as
    SODARUserAdditionalEmail objects are included. If a user has no main email
    set but additional emails exist, the latter are returned.

    :param user: User object
    :return: list
    """
    ret = []
    if user.email:
        ret.append(user.email)
    for e in SODARUserAdditionalEmail.objects.filter(
        user=user, verified=True
    ).order_by('email'):
        ret.append(e.email)
    return ret


def get_project_modify_recipients(
    recipients: list[SODARUser],
) -> list[SODARUser]:
    """
    Filter recipient list for project modify emails. Excludes users with
    notify_email_project set False.

    :param recipients: List of SODARUser objects
    :return: List of SODARUser objects
    """
    return [
        u
        for u in recipients
        if app_settings.get(APP_NAME, 'notify_email_project', user=u)
    ]


def send_mail(
    subject: str,
    message: str,
    recipient_list: list,
    request: Optional[HttpRequest] = None,
    reply_to: Optional[list] = None,
    cc: Optional[list] = None,
    bcc: Optional[list] = None,
) -> int:
    """
    Wrapper for send_mail() with logging and error messaging.

    :param subject: Message subject (string)
    :param message: Message body (string)
    :param recipient_list: Recipients of email (list)
    :param request: HttpRequest object (optional)
    :param reply_to: List of emails for the "reply-to" header (optional)
    :param cc: List of emails for "cc" field (optional)
    :param bcc: List of emails for "bcc" field (optional)
    :return: Amount of sent email (int)
    """
    try:
        e = EmailMessage(
            subject=subject,
            body=message,
            from_email=EMAIL_SENDER,
            to=recipient_list,
            reply_to=reply_to if isinstance(reply_to, list) else [],
            cc=cc if isinstance(cc, list) else [],
            bcc=bcc if isinstance(bcc, list) else [],
        )
        ret = e.send(fail_silently=False)
        logger.debug(
            '{} email{} sent to {}'.format(
                ret, 's' if ret != 1 else '', ', '.join(recipient_list)
            )
        )
        return ret
    except Exception as ex:
        error_msg = f'Error sending email: {str(ex)}'
        logger.error(error_msg)
        if DEBUG:
            raise ex
        if request:
            messages.error(request, error_msg)
        return 0


# Sending functions ------------------------------------------------------------


def send_role_change_mail(
    change_type: str,
    project: Project,
    user: SODARUser,
    role: Role,
    request: HttpRequest,
) -> int:
    """
    Send email to user when their role in a project has been changed.

    :param change_type: Change type ('create', 'update', 'delete')
    :param project: Project object
    :param user: User object
    :param role: Role object (can be None for deletion)
    :param request: HttpRequest object
    :return: Amount of sent email (int)
    """
    subject = get_role_change_subject(change_type, project)
    message = get_role_change_body(
        change_type=change_type,
        project=project,
        user_name=user.get_full_name(),
        role_name=role.name if role else '',
        issuer=request.user,
        request=request,
    )
    issuer_emails = get_user_addr(request.user)
    return send_mail(
        subject, message, get_user_addr(user), request, issuer_emails
    )


def send_project_leave_mail(
    project: Project, user: SODARUser, request: Optional[HttpRequest] = None
) -> int:
    """
    Send email to project owners and delegates when a user leaves a project.

    :param project: Project object
    :param user: User object
    :param request: HttpRequest object or None
    :return: Amount of sent email (int)
    """
    p_label = get_display_name(project.type)
    subject = SUBJECT_PREFIX + SUBJECT_ROLE_LEAVE.format(
        user_name=user.username, project_label=p_label, project=project.title
    )
    message = MESSAGE_ROLE_LEAVE.format(
        user_name=user.username, project_label=p_label, project=project.title
    )
    mail_count = 0
    recipients = [
        a.user
        for a in project.get_roles(max_rank=ROLE_RANKING[PROJECT_ROLE_DELEGATE])
        if a.user != user
        and a.user.is_active
        and app_settings.get(APP_NAME, 'notify_email_role', user=a.user)
    ]
    for r in recipients:
        mail_count += send_mail(subject, message, get_user_addr(r), request)
    return mail_count


def send_invite_mail(invite: ProjectInvite, request: HttpRequest) -> int:
    """
    Send an email invitation to user not yet registered in the system.

    :param invite: ProjectInvite object
    :param request: HttpRequest object
    :return: Amount of sent email (int)
    """
    message = get_invite_body(
        project=invite.project,
        issuer=invite.issuer,
        role_name=invite.role.name,
        invite_url=invite.get_url(request),
        date_expire_str=localtime(invite.date_expire).strftime(
            '%Y-%m-%d %H:%M'
        ),
    )
    message += get_invite_message(invite.message)
    message += get_email_footer(request, settings_link=False)
    subject = get_invite_subject(invite.project)
    issuer_emails = get_user_addr(invite.issuer)
    return send_mail(subject, message, [invite.email], request, issuer_emails)


def send_invite_accept_mail(
    invite: ProjectInvite, request: HttpRequest, user: SODARUser
) -> int:
    """
    Send a notification email to the issuer of an invitation when a user
    accepts the invitation.

    :param invite: ProjectInvite object
    :param request: HttpRequest object
    :param user: User invited
    :return: Amount of sent email (int)
    """
    subject = SUBJECT_PREFIX + SUBJECT_ACCEPT.format(
        user=get_email_user(user),
        project_label=get_display_name(invite.project.type),
        project=invite.project.title,
    )
    message = get_email_header(
        MESSAGE_HEADER.format(
            recipient=invite.issuer.get_full_name(), site_title=SITE_TITLE
        )
    )
    message += MESSAGE_ACCEPT_BODY.format(
        role=invite.role.name,
        project=invite.project.title,
        user=get_email_user(user),
        site_title=SITE_TITLE,
        project_label=get_display_name(invite.project.type),
    )

    if not settings.PROJECTROLES_EMAIL_SENDER_REPLY:
        message += NO_REPLY_NOTE
    message += get_email_footer(request)
    return send_mail(subject, message, get_user_addr(invite.issuer), request)


def send_invite_expiry_mail(
    invite: ProjectInvite, request: HttpRequest, user_name: str
) -> int:
    """
    Send a notification email to the issuer of an invitation when a user
    attempts to accept an expired invitation.

    :param invite: ProjectInvite object
    :param request: HttpRequest object
    :param user_name: User name of invited user
    :return: Amount of sent email (int)
    """
    subject = SUBJECT_PREFIX + SUBJECT_EXPIRY.format(
        user_name=user_name, project=invite.project.title
    )
    message = get_email_header(
        MESSAGE_HEADER.format(
            recipient=invite.issuer.get_full_name(), site_title=SITE_TITLE
        )
    )
    message += MESSAGE_EXPIRY_BODY.format(
        role=invite.role.name,
        project=invite.project.title,
        user_name=user_name,
        user_email=invite.email,
        date_expire=localtime(invite.date_expire).strftime('%Y-%m-%d %H:%M'),
        site_title=SITE_TITLE,
        project_label=get_display_name(invite.project.type),
    )

    if not settings.PROJECTROLES_EMAIL_SENDER_REPLY:
        message += NO_REPLY_NOTE
    message += get_email_footer(request)
    return send_mail(subject, message, get_user_addr(invite.issuer), request)


def send_project_create_mail(
    project: Project, recipients: list[SODARUser], request: HttpRequest
) -> int:
    """
    Send email about project creation to owners and delegates of the parent
    category, excluding the project creator. Also excludes users who have set
    notify_email_role to False.

    Received recipient list must be filtered for email preferences.

    :param project: Project object
    :param recipients: List of SODARUser objects
    :param request: HttpRequest object
    :return: Amount of sent email (int)
    """
    parent = project.parent
    if not parent:
        return 0
    recipients = get_project_modify_recipients(recipients)
    if not recipients:
        return 0
    owner = project.get_owner().user

    subject = SUBJECT_PREFIX + SUBJECT_PROJECT_CREATE.format(
        project_type=get_display_name(project.type, title=True),
        project=project.title,
        user=get_email_user(request.user),
    )
    mail_count = 0
    for r in recipients:
        message = get_email_header(
            MESSAGE_HEADER.format(
                recipient=r.get_full_name(), site_title=SITE_TITLE
            )
        )
        message += MESSAGE_PROJECT_CREATE_BODY.format(
            user=get_email_user(request.user),
            project_type=get_display_name(project.type),
            category=parent.title,
            project=project.title,
            owner=get_email_user(owner),
            project_url=request.build_absolute_uri(
                reverse(
                    'projectroles:detail',
                    kwargs={'project': project.sodar_uuid},
                )
            ),
        )
        message += get_email_footer(request)
        mail_count += send_mail(
            subject,
            message,
            get_user_addr(r),
            request,
            reply_to=get_user_addr(request.user),
        )
    return mail_count


def send_project_move_mail(
    project: Project, recipients: list[SODARUser], request: HttpRequest
) -> int:
    """
    Send email about project being moved to the owners and delegates of the
    parent category, excluding the project creator. Also excludes users who have
    set notify_email_role to False.

    Received recipient list must be filtered for email preferences.

    :param project: Project object
    :param recipients: List of SODARUser objects
    :param request: HttpRequest object
    :return: Amount of sent email (int)
    """
    parent = project.parent
    if not parent:
        return 0
    recipients = get_project_modify_recipients(recipients)
    if not recipients:
        return 0
    owner = project.get_owner().user

    subject = SUBJECT_PREFIX + SUBJECT_PROJECT_MOVE.format(
        project_type=get_display_name(project.type, title=True),
        project=project.title,
        user=get_email_user(request.user),
    )
    mail_count = 0
    for r in recipients:
        message = get_email_header(
            MESSAGE_HEADER.format(
                recipient=r.get_full_name(), site_title=SITE_TITLE
            )
        )
        message += MESSAGE_PROJECT_MOVE_BODY.format(
            user=get_email_user(request.user),
            project_type=get_display_name(project.type),
            category=parent.title,
            project=project.title,
            owner=get_email_user(owner),
            project_url=request.build_absolute_uri(
                reverse(
                    'projectroles:detail',
                    kwargs={'project': project.sodar_uuid},
                )
            ),
        )
        message += get_email_footer(request)
        mail_count += send_mail(
            subject,
            message,
            get_user_addr(r),
            request,
            reply_to=get_user_addr(request.user),
        )
    return mail_count


def send_project_archive_mail(
    project: Project, action: str, request: HttpRequest
) -> int:
    """
    Send a notification email on project archiving or unarchiving.

    :param project: Project object
    :param action: String ("archive" or "unarchive")
    :param request: HttpRequest object
    :return: Amount of sent email (int)
    """
    recipients = [
        a.user
        for a in project.get_roles()
        if a.user != request.user
        and a.user.is_active
        and app_settings.get(APP_NAME, 'notify_email_project', user=a.user)
    ]
    if not recipients:
        return 0
    mail_count = 0

    if action == 'archive':
        subject = SUBJECT_PROJECT_ARCHIVE
        body = MESSAGE_PROJECT_ARCHIVE
    else:
        subject = SUBJECT_PROJECT_UNARCHIVE
        body = MESSAGE_PROJECT_UNARCHIVE
    subject = SUBJECT_PREFIX + subject.format(
        project_label_title=get_display_name(project.type, title=True),
        project=project.title,
        user=request.user.get_full_name(),
    )
    body_final = body.format(
        project_label_title=get_display_name(project.type, title=True),
        project_label=get_display_name(project.type),
        project=project.title,
        user=request.user.get_full_name(),
        project_url=request.build_absolute_uri(
            reverse(
                'projectroles:detail', kwargs={'project': project.sodar_uuid}
            )
        ),
    )

    for r in recipients:
        message = get_email_header(
            MESSAGE_HEADER.format(
                recipient=r.get_full_name(), site_title=SITE_TITLE
            )
        )
        message += body_final
        if not settings.PROJECTROLES_EMAIL_SENDER_REPLY:
            message += NO_REPLY_NOTE
        message += get_email_footer(request)
        mail_count += send_mail(subject, message, get_user_addr(r), request)
    return mail_count


def send_project_delete_mail(
    project: Project, recipients: list[SODARUser], request: HttpRequest
) -> int:
    """
    Send a notification email on project deletion.

    Received recipient list must be filtered for email preferences.

    :param project: Project object
    :param recipients: List of SODARUser objects
    :param request: HttpRequest object
    :return: Amount of sent email (int)
    """
    user = request.user
    recipients = [
        u
        for u in recipients
        if app_settings.get(APP_NAME, 'notify_email_project', user=u)
    ]
    if not recipients:
        return 0

    mail_count = 0
    subject = SUBJECT_PROJECT_DELETE
    body = MESSAGE_PROJECT_DELETE
    subject = SUBJECT_PREFIX + subject.format(
        project_label_title=get_display_name(project.type, title=True),
        project=project.title,
        user=user.get_full_name(),
    )
    body_final = body.format(
        project_label=get_display_name(project.type),
        project=project.title,
        user=user.get_full_name(),
    )

    for r in recipients:
        message = get_email_header(
            MESSAGE_HEADER.format(
                recipient=r.get_full_name(), site_title=SITE_TITLE
            )
        )
        message += body_final
        if not settings.PROJECTROLES_EMAIL_SENDER_REPLY:
            message += NO_REPLY_NOTE
        message += get_email_footer(request)
        mail_count += send_mail(subject, message, get_user_addr(r), request)
    return mail_count


def send_generic_mail(
    subject_body: str,
    message_body: str,
    recipient_list: list,
    request: Optional[HttpRequest] = None,
    reply_to: Optional[list] = None,
    cc: Optional[list] = None,
    bcc: Optional[list] = None,
    settings_link: bool = True,
) -> int:
    """
    Send a generic mail with standard header and footer and no-reply
    notifications.

    :param subject_body: Subject body without prefix (string)
    :param message_body: Message body before header or footer (string)
    :param recipient_list: Recipients (list of User objects or email strings)
    :param request: HttpRequest object (optional)
    :param reply_to: List of emails for the "reply-to" header (optional)
    :param cc: List of emails for "cc" field (optional)
    :param bcc: List of emails for "bcc" field (optional)
    :param settings_link: Include link to user settings if True (optional)
    :return: Amount of mail sent (int)
    """
    subject = SUBJECT_PREFIX + subject_body
    ret = 0
    for recipient in recipient_list:
        if isinstance(recipient, User):
            recp_name = recipient.get_full_name()
            recp_addr = get_user_addr(recipient)
        else:
            recp_name = 'recipient'
            recp_addr = [recipient]
        message = get_email_header(
            MESSAGE_HEADER.format(recipient=recp_name, site_title=SITE_TITLE)
        )
        message += message_body
        if not reply_to and not settings.PROJECTROLES_EMAIL_SENDER_REPLY:
            message += NO_REPLY_NOTE
        message += get_email_footer(request, settings_link)
        ret += send_mail(
            subject, message, recp_addr, request, reply_to, cc, bcc
        )
    return ret
